/** Derivative Checker
 *
 *  \file   derivativeChecker.cpp
 *  \author Benedict Geihe
 *  \date   01.01.2013
 *
 */



#include "derivativeChecker.h"

#include <string>
#include <sstream>
#include <cassert>
#include <cstdlib>
#include <fstream>

#if defined(__MINGW32_VERSION) || defined(_MSC_VER) || defined(__BCPLUSPLUS__)
#include <process.h>
#define GNUPLOT_PATH string("\"C:\\Program Files\\gnuplot\\bin\\wgnuplot.exe\" ")
#else
#include <unistd.h>
#define GNUPLOT_PATH string("gnuplot ")
#endif

//! Konstruktor
DerivativeChecker::DerivativeChecker(const Operator<Vector,double> & energy, const Operator<Vector,Vector> & gradient)
    : _energy ( energy )
    , _gradient ( gradient )
    , _delta ( 1e-3 )
    , _samples ( 20 )
{}
// ------------------------------------------------------------------------------------------------

//! Check single direction
void DerivativeChecker::check( const unsigned int dof, const Vector & point, string filename )
{
    assert ( dof < point.size() );

    Vector direction( point.size() );
    direction.setZero();
    direction[dof]=1.;

    Vector currentpoint( point );
    currentpoint.addMultiple( -_delta * _samples/2, direction );

    Vector energySamples(_samples);
    Vector xvalues(_samples);
    double currentx = point[dof] - _delta * _samples/2;
    for ( int i = 0; i < _samples; ++i )
    {
        xvalues[i] = currentx;
        _energy.apply( currentpoint, energySamples[i] );
        currentx += _delta;
        currentpoint.addMultiple( _delta, direction );
    }

    Vector gradient ( point.size() );
    _gradient.apply( point, gradient );


    // plot
    stringstream output;
    output << filename.substr( 0, filename.rfind(".") ) << setw(5) << setfill('0') << dof << ".png";
    filename = output.str();

    output.str("");
    output << filename.substr( 0, filename.rfind(".") ) << "_" << getpid();
    string tmpfile = ".gp_" + output.str() + ".tmp";
    fstream file;
    file.open( tmpfile.c_str(), fstream::out );

    if ( file.is_open() )
    {
        file.precision(100);
        file << "set terminal png large size 800,800 enhanced" << endl;
        file << "set output \"" << filename << "\"" << endl;
        file << "set xtics nomirror" << endl;
        file << "set ytics nomirror" << endl;
        file << "plot " << energySamples[_samples/2] << " + " << gradient[dof] << " * (x-" << xvalues[_samples/2] << ")  title 'gradient' w l, '-' title 'energy' w l" << endl;
        for ( int i = 0; i < _samples; ++i )
        {
            file << xvalues[i] << " " << energySamples[i] << endl;
        }
        file << "e" << endl;
        file.close();
        if ( system( (GNUPLOT_PATH + tmpfile).c_str() ) != 0 ) throw "!!! system call failed !!!";
        remove( tmpfile.c_str() );
    }
}
// ------------------------------------------------------------------------------------------------

//! Check all directions
void DerivativeChecker::checkAll( const Vector & point, string filename )
{
    cerr << endl;
    for ( unsigned int i = 0; i < point.size(); ++i )
    {
      cerr << "\rChecking " << 0 << "/" << point.size()-1;
      check(i,point,filename);
    }
    cerr << endl << endl;
}
// ------------------------------------------------------------------------------------------------
