/** Gradient descent
 *  Numerical Algorithms (Prof. M. Rumpf) WS 2012/13
 *
 *  \file   gradientDescent.cpp
 *  \author Benedict Geihe, Behrend Heeren
 *  \date   15.01.2013
 *
 */


#include "gradientDescent.h"

//! Konstruktor
GradientDescent::GradientDescent(const Operator<Vector,double> & energy, const Operator<Vector,Vector> & gradient)
    : _energy ( energy )
    , _gradient ( gradient )
    , _TOL ( 1e-8 )
    , _ETOL ( 1e-8 )
    , _maxIter ( 10000 )
    , _armijoBeta ( 0.5 )
    , _minStepsize ( 1e-10 )
    , _maxStepsize ( 1e10 )
    , _verbose ( true )
    , _useQuadraticApproximation( false )
{}
// ------------------------------------------------------------------------------------------------

//! \brief Implementierung der applyAdd-Methode vom FEOperator
//! Startet den Gradientenabstieg mit Startwert \f$ x_0 \f$
void GradientDescent::apply(const Vector & x_0, Vector & x_final, double lambda)
{
    Vector x(x_0);
    init();
    while ( performSingleStep(x) == NotTerminated );
    checkTermination();
    x_final.setToMultiple(lambda, x);
}
// ------------------------------------------------------------------------------------------------

//! Eigentlicher Gradientenabstieg
void GradientDescent::init()
{
    _lastStepsize = 1.;
    _currentIter = 0;
    _lastTerminationReason = NotTerminated;
    printHeader();
}
// ------------------------------------------------------------------------------------------------

//! Eigentlicher Gradientenabstieg
void GradientDescent::checkTermination() const
{
    if (_verbose)
        switch (_lastTerminationReason)
        {
            case NoStepsizeFound:
                cout << "*** No stepsize found." << endl;
                break;
            case MaxIterReached:
                cout << "*** Maximum number of iterations (" << _maxIter << ") reached." << endl;
                break;
            case TOL_Reached:
                cout << "*** Gradient tolerance (" << _TOL << ") reached." << endl;
                break;
            case ETOL_Reached:
                cout << "*** Energy decrease tolerance (" << _ETOL << ") reached." << endl;
                break;
            case NotTerminated:
              cout << "*** Not terminated yet!" << endl;
              break;
          }
}
// ------------------------------------------------------------------------------------------------

//! Eigentlicher Gradientenabstieg
TerminationReason GradientDescent::performSingleStep(Vector & x)
{
    TerminationReason tr = NotTerminated;
    ++_currentIter;

    Vector gradient(x.size());

    // Gradienten bestimmen
    _gradient.apply(x, gradient );

    Vector update( gradient );
    update *= -1.;

    double gradNorm = update.norm_l2();

    // Schrittweite bestimmen
    double stepsize = findStepsizeWithArmijo(x, update, gradient );
    if (stepsize == 0.)
      tr = NoStepsizeFound;
    else
    {
      // neues x berechnen
      x.addMultiple(stepsize, update);

      // Schritt ausgeben
      printIterate(gradNorm, _energyNew);

      // Abbruchkriterien testen
      if (_currentIter == _maxIter)
        tr = MaxIterReached;
      if (gradNorm < _TOL)
        tr = TOL_Reached;
      if ( _energyOld - _energyNew < _ETOL)
        tr = ETOL_Reached;
    }

    if ( tr != NotTerminated ) _lastTerminationReason = tr;
    return tr;
}
// ------------------------------------------------------------------------------------------------

//! Schrittweitensteuerung mit Armijo-Regel
double GradientDescent::findStepsizeWithArmijo(const Vector & x, const Vector & p, const Vector & grad )
{
    Vector x_neu(x.size());
    _energy.apply(x, _energyOld);

    double stepsize = _lastStepsize;

    // neues x und Funktionswert dort berechnen
    x_neu = x;
    x_neu.addMultiple(stepsize, p);
    _energy.apply(x_neu, _energyNew);
    _evaluations = 2;

    // quadratic approximation
    if (_useQuadraticApproximation)
    {
        double dsqr = grad * p;
        stepsize = -dsqr*stepsize*stepsize*.5 / ( _energyNew - _energyOld - dsqr * stepsize );
        x_neu = x;
        x_neu.addMultiple(stepsize, p);
        _energy.apply(x_neu, _energyNew);
        ++_evaluations;
    }

    // erwartete und tatsaechliche Steigung berechnen
    double erwarteteSteigung = grad * p;
    double tatsSteigung = (_energyNew - _energyOld) / stepsize;

    if (tatsSteigung > _armijoBeta * erwarteteSteigung)
    {
        // Schrittweite verkleinern
        do
        {
            stepsize *= 0.5;
            if (stepsize < _minStepsize)
                return 0.;
            x_neu = x;
            x_neu.addMultiple(stepsize, p);
            _energy.apply(x_neu, _energyNew);
            ++_evaluations;
            tatsSteigung = (_energyNew - _energyOld) / stepsize;
        } while (tatsSteigung > _armijoBeta * erwarteteSteigung);
    }
    else
    {
        // Schrittweite vergroessern
        double energyBefore;
        do
        {
            energyBefore = _energyNew;
            if (stepsize > _maxStepsize)
                return _lastStepsize = _maxStepsize;
            stepsize *= 2.;
            x_neu = x;
            x_neu.addMultiple(stepsize, p);
            _energy.apply(x_neu, _energyNew);
            ++_evaluations;
            tatsSteigung = (_energyNew - _energyOld) / stepsize;
        } while (tatsSteigung < _armijoBeta * erwarteteSteigung);
        // es ist einmal zu oft vergroessert worden:
        stepsize *= 0.5;
        _energyNew = energyBefore;
    }
    return _lastStepsize = stepsize;
}
// ------------------------------------------------------------------------------------------------

//! Ausgabe des Kopfes im verbose-Modus
void GradientDescent::printHeader() const
{
    cout << "iteration:    "
         << setw(20) << left << "step size:"
         << setw(20) << "norm gradient:"
         << setw(20) << "energy:"
         << setw(20) << "evaluations:" << endl
         << "--------------------------------------------------------------------------------------" << endl;
}
// ------------------------------------------------------------------------------------------------

//! Ausgabe des aktuellen Schrittes im verbose-Modus
void GradientDescent::printIterate(double gradNorm, double energy ) const
{
    cout << setw(10) << right << _currentIter << "    " << left
         << setw(20) << _lastStepsize
         << setw(20) << gradNorm
         << setw(20) << energy
         << setw(20) << _evaluations
         << endl;
}
// ------------------------------------------------------------------------------------------------

//! Schaltet den verbose-Modus ein/aus
void GradientDescent::setVerbose(bool verbose)
{
    _verbose = verbose;
}
// ------------------------------------------------------------------------------------------------

//! Schaltet quadratische Approximation in der Schrittweitensteuerung ein/aus
void GradientDescent::setQuadraticApproximation(bool quad)
{
    _useQuadraticApproximation = quad;
}
// ------------------------------------------------------------------------------------------------
