/** Elastic problem
 *
 *  \file   ElasticProblem.cpp
 *  \author Benedict Geihe, Behrend Heeren
 *  \date   21.01.2013
 *
 */


#include "ElasticProblem.h"

//! constructor
ElasticProblem::ElasticProblem( const double eps, ScenarioType scenario )
{
    // generate the grid
    switch ( scenario )
    {
        case SQUARE:
        {
            _grid = Grid::createRectangularGrid(0.,0.,1.,1.,10);
            break;
        }
        case CANTILEVER:
        {
            _grid = Grid( Node(0.0,0.0), Node(1.0,0.0), Node(0.0,1.0) );
            _grid.addNodeAndTriangle( Node(1.0,1.0), Grid::DartIterator( _grid.getTriangle(0), 0, 1 ) );
            _grid.addNodeAndTriangle( Node(2.0,0.0), Grid::DartIterator( _grid.getTriangle(1), 2, 0 ) );
            _grid.addNodeAndTriangle( Node(2.0,1.0), Grid::DartIterator( _grid.getTriangle(2), 2, 0 ) );
            for ( int i = 0; i < 5; ++i ) _grid.refine4All();
            break;
        }
        case XMAS:
        {
            generateMysteriousGrid( _grid );
            break;
        }
    }
    _N = _grid.getNumberOfNodes();
    _rhs.resize(2*_N);
    _u.resize(2*_N);
    _pf.resize(_N);

    // generate phase field
    switch ( scenario )
    {
        case SQUARE:
        {
            generateCircularPhasefield ( _pf, _grid, SmallVector2<>(.5,.5), .4, eps );
            break;
        }
        case CANTILEVER:
        {
            generateBarPhasefield( _pf, _grid, .01, .5, eps );
            break;
        }
        case XMAS:
        {
            generateMysteriousPhasefield( _pf, _grid, eps );
            break;
        }
    }

    // set loads
    switch ( scenario )
    {
        case SQUARE:
        {
            _load[0] =  0.1;
            _load[1] =  0.0;
            break;
        }
        case CANTILEVER:
        {
            _load[0] =  0.0;
            _load[1] = -0.1;
            break;
        }
        case XMAS:
        {
            _load[0] =  0.1;
            _load[1] = -0.1;
            break;
        }
    }

    // prepare boundary data
    for ( Grid::BoundaryConstIterator bIter = _grid; bIter.notAtEnd(); ++bIter )
    {
        switch ( scenario )
        {
            case SQUARE:
            {
                if (( fabs( bIter.getStart()[1] ) > 1.-1e-14 ) && ( fabs ( bIter.getEnd()[1] ) > 1.-1e-14 ))
                {
                    // integrate linear finite element times constant loading explicitly
                    _rhs[      bIter.getStartIndex() ] += bIter.getLength() * .5 * _load[0];
                    _rhs[      bIter.getEndIndex()   ] += bIter.getLength() * .5 * _load[0];
                    _rhs[ _N + bIter.getStartIndex() ] += bIter.getLength() * .5 * _load[1];
                    _rhs[ _N + bIter.getEndIndex()   ] += bIter.getLength() * .5 * _load[1];
                }
                // Clamped on bottom
                if ( fabs( bIter.getStart()[1] ) < 1e-14 )
                {
                    _bdryMask.mask(      bIter.getStartIndex() );
                    _bdryMask.mask( _N + bIter.getStartIndex() );
                }
                break;
            }
            case CANTILEVER:
            {
                if (( fabs( bIter.getStart()[0] ) > 2.-1e-14 ) && (( fabs ( bIter.getStart()[1] - .5 ) < 1e-14 ) || ( fabs ( bIter.getEnd()[1] - .5 ) < 1e-14 )))
                {
                    // integrate linear finite element times constant loading explicitly
                    _rhs[      bIter.getStartIndex() ] += bIter.getLength() * .5 * _load[0];
                    _rhs[      bIter.getEndIndex()   ] += bIter.getLength() * .5 * _load[0];
                    _rhs[ _N + bIter.getStartIndex() ] += bIter.getLength() * .5 * _load[1];
                    _rhs[ _N + bIter.getEndIndex()   ] += bIter.getLength() * .5 * _load[1];
                }
                // Clamped on left hand side
                if ( fabs( bIter.getStart()[0] ) < 1e-14 )
                {
                    _bdryMask.mask(      bIter.getStartIndex() );
                    _bdryMask.mask( _N + bIter.getStartIndex() );
                }
                break;
            }
            case XMAS:
            {
                if (  ( ( fabs( bIter.getStart()[1] > 5.5 ) ) || ( fabs( bIter.getEnd()[1] > 5.5 ) ) )
                     && ( fabs( bIter.getStart()[0] < 1e-14 ) ) && ( fabs( bIter.getEnd()[0] < 1e-14 ) ) )
                {
                    // integrate linear finite element times constant loading explicitly
                    _rhs[      bIter.getStartIndex() ] += bIter.getLength() * .5 * _load[0];
                    _rhs[      bIter.getEndIndex()   ] += bIter.getLength() * .5 * _load[0];
                    _rhs[ _N + bIter.getStartIndex() ] += bIter.getLength() * .5 * _load[1];
                    _rhs[ _N + bIter.getEndIndex()   ] += bIter.getLength() * .5 * _load[1];
                }
                // Clamped at bottom
                if ( fabs( bIter.getStart()[1] ) < 1e-14 )
                {
                    _bdryMask.mask(      bIter.getStartIndex() );
                    _bdryMask.mask( _N + bIter.getStartIndex() );
                }
                break;
            }
        }
    }

    // compute solution for current setting
    solve();
}
// ------------------------------------------------------------------------------------------------

//! update the current phasefield, recompute elastic solution if necessary
void ElasticProblem::updatePhasefield( const Vector & pf )
{
    _pf -= pf;
    double diff = _pf.norm_l2();
    _pf = pf;
    if ( diff > 1e-15 ) solve();
}
// ------------------------------------------------------------------------------------------------

//! get solution reference
const Vector & ElasticProblem::getSolution() const
{
    return _u;
}
// ------------------------------------------------------------------------------------------------

//! get phase field reference
const Vector & ElasticProblem::getPhasefield() const
{
    return _pf;
}
// ------------------------------------------------------------------------------------------------

//! get grid reference
const Grid & ElasticProblem::getGrid() const
{
    return _grid;
}
// ------------------------------------------------------------------------------------------------


//! solve the elastic problem
void ElasticProblem::solve()
{
    // create the elastic bilinear form
    PhasefieldElastStiffBilf elastBilf ( _pf );
    // create the vector valued linear finite element operator
    LinearFEOperator<PhasefieldElastStiffBilf, BarQuad1, 2> elastStiffOp(_grid, elastBilf);
    // apply boundary mask
    _bdryMask.apply( elastStiffOp );

    // solve by CG
    Solver<SparseMatrix> solver( elastStiffOp, false );
    solver.setMaxIter( 50000 );
    solver.setTol( 1E-12 );
    solver.CGapply( _rhs, _u );
}
// ------------------------------------------------------------------------------------------------

//! initialize a circular phase field
void ElasticProblem::generateCircularPhasefield ( Vector & pf, const Grid & grid, const SmallVector2<> & mid, const double radius, const double eps )
{
  for (Grid::NodeConstIterator nIter = grid; nIter.notAtEnd(); ++nIter)
  {
    SmallVector2<> dist ( mid );
    dist -= (*nIter);
    pf[ nIter.getIndex() ] = tanh ( 3./4. * ( radius - dist.norm_l2() ) / eps );
  }
}
// ------------------------------------------------------------------------------------------------

//! initialize a bar phase field
void ElasticProblem::generateBarPhasefield ( Vector & pf, const Grid & grid, const double width, const double y, const double eps )
{
  for (Grid::NodeConstIterator nIter = grid; nIter.notAtEnd(); ++nIter)
  {
    pf[ nIter.getIndex() ] = tanh ( 3./4. * ( width - fabs( (*nIter)[1]-y ) ) / eps );
  }
}
// ------------------------------------------------------------------------------------------------

//! initialize grid of exercice 7
void ElasticProblem::generateMysteriousGrid( Grid & grid )
{
    vector<Node> nodes;
    nodes.push_back(Node(-1., 0.));
    nodes.push_back(Node( 1., 0.));
    nodes.push_back(Node( 0., 1.));
    nodes.push_back(Node(-3., 2.));
    nodes.push_back(Node(-1., 2.));
    nodes.push_back(Node( 1., 2.));
    nodes.push_back(Node( 3., 2.));
    nodes.push_back(Node(-2., 4.));
    nodes.push_back(Node(-1., 4.));
    nodes.push_back(Node( 0., 4.));
    nodes.push_back(Node( 1., 4.));
    nodes.push_back(Node( 2., 4.));
    nodes.push_back(Node(-1.25, 5.5));
    nodes.push_back(Node(-0.5, 5.5));
    nodes.push_back(Node( 0.5, 5.5));
    nodes.push_back(Node( 1.25, 5.5));
    nodes.push_back(Node( 0., 6.75));

    grid = Grid(nodes[0], nodes[1], nodes[2]);
    grid.addNodeAndTriangle ( nodes[ 4], Grid::DartIterator ( grid.getTriangle( 0), 1, 0 ) );
    grid.addNodeAndTriangle ( nodes[ 5], Grid::DartIterator ( grid.getTriangle( 1), 1, 2 ) );
    grid.addTriangle ( Grid::DartIterator ( grid.getTriangle(2), 2 ), Grid::DartIterator ( grid.getTriangle(0), 0 ) );
    grid.addNodeAndTriangle ( nodes[ 9], Grid::DartIterator ( grid.getTriangle( 2), 1, 0 ) );
    grid.addNodeAndTriangle ( nodes[ 8], Grid::DartIterator ( grid.getTriangle( 4), 1, 0 ) );
    grid.addNodeAndTriangle ( nodes[ 3], Grid::DartIterator ( grid.getTriangle( 5), 1, 0 ) );
    grid.addNodeAndTriangle ( nodes[10], Grid::DartIterator ( grid.getTriangle( 4), 2, 0 ) );
    grid.addNodeAndTriangle ( nodes[ 6], Grid::DartIterator ( grid.getTriangle( 7), 1, 0 ) );
    grid.addNodeAndTriangle ( nodes[13], Grid::DartIterator ( grid.getTriangle( 5), 2, 0 ) );
    grid.addNodeAndTriangle ( nodes[ 7], Grid::DartIterator ( grid.getTriangle( 9), 2, 0 ) );
    grid.addNodeAndTriangle ( nodes[14], Grid::DartIterator ( grid.getTriangle( 7), 2, 0 ) );
    grid.addNodeAndTriangle ( nodes[11], Grid::DartIterator ( grid.getTriangle(11), 2, 0 ) );
    grid.addTriangle ( Grid::DartIterator ( grid.getTriangle(9), 1 ), Grid::DartIterator ( grid.getTriangle(11), 1 ) );
    grid.addNodeAndTriangle ( nodes[16], Grid::DartIterator ( grid.getTriangle(13), 0, 1 ) );
    grid.addNodeAndTriangle ( nodes[12], Grid::DartIterator ( grid.getTriangle(14), 2, 0 ) );
    grid.addNodeAndTriangle ( nodes[15], Grid::DartIterator ( grid.getTriangle(14), 1, 2 ) );

    bool done = false;
    do
    {
        done = true;
        for (Grid::TriangleIterator tIter = grid; tIter.notAtEnd(); ++tIter)
        {
            if ( tIter->getLongestEdgeLength() > .2 )
            {
                tIter->markForRefinement();
                done = false;
            }
        }
        grid.refineMarkedTriangles();
    } while (!done);
}
// ------------------------------------------------------------------------------------------------

//! initialize phase field of exercise 7
void ElasticProblem::generateMysteriousPhasefield( Vector & pf, const Grid & grid, const double eps )
{
    SmallVector2<> diff;
    double dist;
    vector< SmallVector2<> > mids;
    vector< double > radius;
    mids.push_back( SmallVector2<>(-2.  ,2.4) );
    radius.push_back(.3);
    mids.push_back( SmallVector2<>( 2.  ,2.4) );
    radius.push_back(.3);
    mids.push_back( SmallVector2<>(- .9 ,3.1) );
    radius.push_back(.25);
    mids.push_back( SmallVector2<>(  .9 ,3.1) );
    radius.push_back(.25);
    mids.push_back( SmallVector2<>(-1.25,4.3) );
    radius.push_back(.2);
    mids.push_back( SmallVector2<>( 1.25,4.3) );
    radius.push_back(.2);
    mids.push_back( SmallVector2<>(- .8 ,5.7) );
    radius.push_back(.1);
    mids.push_back( SmallVector2<>(  .8 ,5.7) );
    radius.push_back(.1);
    mids.push_back( SmallVector2<>( 0.  ,6.2) );
    radius.push_back(.2);

    for (Grid::NodeConstIterator nIter = grid; nIter.notAtEnd(); ++nIter)
    {
        if ( (*nIter)[1] < 5. )
            dist = fabs( (*nIter)[0] );
        else  {
            diff[0] = 0.; diff[1] = 5.;
            diff -= (*nIter);
            dist = diff.norm_l2();
        }
        dist -= .2;
        for ( unsigned int i = 0; i < mids.size(); ++i )
        {
            diff = mids[i];
            diff -= (*nIter);
            dist = std::min( dist, diff.norm_l2()-radius[i] );
        }
        pf[ nIter.getIndex() ] = tanh ( -3./4. * dist / eps );
    }
}
// ------------------------------------------------------------------------------------------------
