/** Linearer Finite-Elemente-Operator fuer das Programmierpraktikum
 *
 *  \file   LinearFEOperator.cpp
 *  \author Stefan W. von Deylen, Benedict Geihe
 *  \date   01.10.2008
 *
 */

#include "LinearFEOperator.h"
#include "bilf.h"
#include "diffusiveBilf.h"
#include "twoscaleBilf.h"
#include "elastBilf.h"

/***************************************************************************************************/
/*                                      LinearFEOperator                                           */
/***************************************************************************************************/
//! Konstruktor mit Bilinearform
template <typename BilfType, typename QuadType, int Dim>
LinearFEOperator< BilfType, QuadType, Dim >::LinearFEOperator( const Grid & grid, BilfType & bilf )
    : SparseMatrix( Dim*grid.getNumberOfNodes(), Dim*grid.getNumberOfNodes() )
    , _grid ( grid )
    , _bilf( bilf )
{
    Matrix locMat(3,3);
    GlobalIndex global[3];
    int N = grid.getNumberOfNodes(); // dofs per dimension

    for ( Grid::TriangleConstIterator iter ( _grid ); iter.notAtEnd(); ++iter )
    {
        // get the global node indices of the current triangle
        for ( int i = 0; i < 3; ++i ) {
            global[ i ] = iter->getNodeIndex( i );
        }

        // for vector valued problems loop over all combinations of components
        for ( int k = 0; k < Dim; ++k )  {
            for ( int l = 0; l < Dim; ++l )  {
                // assemble the local matrix for the current element
                assembleLocalMatrix ( *iter, locMat, k, l );

                // finally add the locally computed values to the matrix
                for ( int i = 0; i < 3; ++i ) {
                    for ( int j = 0; j < 3; ++j ) {
                        (*this)( k*N + global[i], l*N + global[j] ) += locMat( i, j );
                    }
                }
            }
        }
    }
}

//! Assembliert lokal auf einem Dreieck
template <typename BilfType, typename QuadType, int Dim>
void LinearFEOperator< BilfType, QuadType, Dim >::assembleLocalMatrix( const Triangle & triangle, Matrix & mat, int k, int l ) const
{
    mat.setZero();

    for ( int i = 0; i < 3; ++i )  {
        for ( int j = 0; j < 3; ++j )  {
            for ( int q = 0; q < QuadType::_numPoints; ++q )  {
                mat(i,j) += QuadType::_weight[q] * BilfEvaluator<BilfType,Dim>::evaluate( _bilf, triangle, QuadType::_barCoord[q], i, j, k, l );
            }
            mat(i,j) *= triangle.getVolume();
        }
    }
}

//! Implementierung der apply-Funktion: verwendet die der SparseMatrix
template <typename BilfType, typename QuadType, int Dim>
void LinearFEOperator< BilfType, QuadType, Dim >::apply( const Vector& x, Vector& b, double lambda ) const
{
    SparseMatrix::apply( x, b, lambda);
}

//! Implementierung der applyAdd-Funktion: verwendet die der SparseMatrix
template <typename BilfType, typename QuadType, int Dim>
void LinearFEOperator< BilfType, QuadType, Dim >::applyAdd( const Vector& x, Vector& b, double lambda ) const
{
    SparseMatrix::applyAdd( x, b, lambda);
}


// explizte Instanzierungen
template class LinearFEOperator<MassBilf, BarQuad1>;
template class LinearFEOperator<MassBilf, BarQuad2>;
template class LinearFEOperator<MassBilf, BarQuad3>;

template class LinearFEOperator<StiffBilf, BarQuad1>;
template class LinearFEOperator<StiffBilf, BarQuad2>;
template class LinearFEOperator<StiffBilf, BarQuad3>;

template class LinearFEOperator<IsoDiffusiveStiffBilf, BarQuad1>;
template class LinearFEOperator<IsoDiffusiveStiffBilf, BarQuad2>;
template class LinearFEOperator<IsoDiffusiveStiffBilf, BarQuad3>;

template class LinearFEOperator<TwoscaleBilf<IsoDiffusiveStiffBilf>, BarQuad1>;

template class LinearFEOperator<PhasefieldElastStiffBilf, BarQuad1, 2>;
