/** Boundary masks fuer das Programmierpraktikum
 *
 *  \file   boundaryMasks.cpp
 *  \author Benedict Geihe, Behrend Heeren
 *  \date   15.10.2012
 *
 */

#include "boundaryMasks.h"

#include <sstream>
#include <cassert>
#include <cstdlib>

#if defined(__MINGW32_VERSION) || defined(_MSC_VER) || defined(__BCPLUSPLUS__)
#include <process.h>
#define GNUPLOT_PATH string("\"C:\\Program Files\\gnuplot\\bin\\wgnuplot.exe\" ")
#else
#include <unistd.h>
#define GNUPLOT_PATH string("gnuplot ")
#endif

/***************************************************************************************************/
/*                                        Dirichlet Mask                                           */
/***************************************************************************************************/
DirichletBoundaryMask::DirichletBoundaryMask()
{}

DirichletBoundaryMask::DirichletBoundaryMask( const Grid& grid ) {
    for (Grid::BoundaryConstIterator bIter = grid; bIter.notAtEnd(); ++bIter)
        _mask.push_back( bIter.getStartIndex() );  
}

void DirichletBoundaryMask::mask ( const int index ) {
    _mask.push_back( index );
}

void DirichletBoundaryMask::apply ( SparseMatrix & matrix ) const {
  for ( unsigned int i = 0; i < _mask.size(); ++i )
    matrix.setRowColToDiagonal( _mask[i] );  
}

void DirichletBoundaryMask::apply ( Vector & vec ) const {
  for ( unsigned int i = 0; i < _mask.size(); ++i )
    vec[ _mask[i] ] = 0.;  
}

/***************************************************************************************************/
/*                                       Periodic Mask                                             */
/***************************************************************************************************/
//! Konstruktor
PeriodicBoundaryMask::PeriodicBoundaryMask( const Grid& grid )
    : _grid ( grid )
{
    // Nutze den Randknoteniterator um alle Randknoten zu sehen
    bool cornerSeen = false;
    int nextCornerIndex = IndexNotSet;
    SmallVector2<> olddiff (1.,1.), newdiff (1.,1.);
    vector<GlobalIndex> nindices, nauxVec;

    for (Grid::BoundaryConstIterator bIter = grid; bIter.notAtEnd(); ++bIter)
    {
      if ( abs( bIter.getStart()[0] - bIter.getStart()[1] ) < 1e-15 ) cornerSeen = true;
        if (cornerSeen) {
            nindices.push_back( bIter.getStartIndex() );
            if ( nextCornerIndex == IndexNotSet )  {
                newdiff = bIter.getStart();
                newdiff-= bIter.getEnd();
                // check for right angle
                if ( abs( newdiff * olddiff ) < 1e-15 ) nextCornerIndex = nindices.size();
                olddiff = newdiff;
            }
        }
        else  {
          nauxVec.push_back( bIter.getStartIndex() );
        }
    }
    nindices.insert( nindices.end(), nauxVec.begin(), nauxVec.end() );

    const unsigned int numBdryNodes = nindices.size();
    assert( numBdryNodes % 2 == 0);
    this->resize( numBdryNodes/2+1 );
    const unsigned int numNodesFirstDir  = nextCornerIndex;
    const unsigned int numNodesSecondDir = this->size()-numNodesFirstDir-1;
    assert( 2*(numNodesFirstDir+numNodesSecondDir) == numBdryNodes );

    for ( unsigned int i = 0; i < this->size(); ++i )
      (*this)[i][0] = nindices[i];

    for ( unsigned int i = 0; i < numNodesFirstDir; ++i )
      (*this)[i][1] = nindices[ numBdryNodes-1-numNodesSecondDir-i ];

    for ( unsigned int i = 0; i < numNodesSecondDir+1; ++i )
      (*this)[numNodesFirstDir+i][1] = nindices[ numBdryNodes-1-i ];

}


void PeriodicBoundaryMask::apply( SparseMatrix & matrix ) const {

  for ( unsigned int i = 0; i < this->size(); ++i ) {
       matrix.addMultipleCol( 1., (*this)[i][0], (*this)[i][1] );
       matrix.addMultipleRow( 1., (*this)[i][0], (*this)[i][1] );
       matrix.setRowColToDiagonal( (*this)[i][0] );
  }
}

void PeriodicBoundaryMask::collapse( Vector & vec ) const {

  for ( unsigned int i = 0; i < this->size(); ++i ) {
       vec[(*this)[i][1]] += vec[(*this)[i][0]];
       vec[(*this)[i][0]] = 0.;
  }
}

void PeriodicBoundaryMask::extend( Vector & vec ) const {

  for ( unsigned int i = this->size(); i > 0; --i ) {
       vec[(*this)[i-1][0]] = vec[(*this)[i-1][1]];
  }
}

