/** Bilinear forms for elasticity problems
 * 
 *  \file   elastBilf.cpp
 *  \author Benedict Geihe, Behrend Heeren
 *  \date   15.12.2012
 *
 */

#include "elastBilf.h"


//! constructor
PhasefieldElastStiffBilf::PhasefieldElastStiffBilf ( const Vector & pf, const double lambda, const double mu, const double weak )
:   _pf ( pf )
,   _lambda_strong ( lambda )
,   _lambda_weak   ( weak )
,   _mu_strong     ( mu )
,   _mu_weak       ( weak )
{}

//! evaluate \f$ ( \lambda \mathrm{tr} \varepsilon[\psi_i^k] \mathds{1} + 2 \mu \varepsilon[\psi_i^k] ) : \varepsilon[\psi_j^l] \f$ on a triangle
double PhasefieldElastStiffBilf::operator() (const Triangle & t, const BarCoord & q, int i, int j, int k, int l ) const
{
    // interpolate linear phasefield at quadrature point
    double chi = t.calcFunctionValue( _pf, q );
    // make characteristic function
    chi += 1.;
    chi *= chi;
    chi *= .25;

    // get gradients
    SmallVector2<> gradi, gradj;
    t.calcCartesianBaseGrad ( i, gradi );
    t.calcCartesianBaseGrad ( j, gradj );

    double ret = 0.;
    // first term: only one component of \psi will be non zero, k,l will give the diagonal entry
    ret += ( chi * _lambda_strong + (1.-chi) * _lambda_weak ) * gradi[k] * gradj[l];
    // second term: off diagonal
    ret += ( chi * _mu_strong + (1.-chi) * _mu_weak ) * gradi[1-k] * gradj[1-l];
    // second term: for identical components diagonal entries meet
    if ( k == l) ret += ( chi * _mu_strong + (1.-chi) * _mu_weak ) * 2. * gradi[k] * gradj[l];

    return ret;
}
