/** Error estimator
 * 
 *  \file   errorEstimator.h
 *  \author Benedict Geihe, Behrend Heeren
 *  \date   19.10.2012
 *
 */

#include "errorEstimator.h"
#include "LinearFEOperator.h"

/**********************************************************************************************/
/*                                   class ErrorEstimator                                     */
/**********************************************************************************************/
ErrorEstimator::ErrorEstimator(Grid & grid, const Vector & f, const Vector & u, double gamma)
    : _grid (grid)
    , _f (f)
    , _u (u)
    , _gamma (gamma)
{}

ErrorEstimator::~ErrorEstimator()
{}

double ErrorEstimator::markTriangles()
{
    Vector errEstimates ( _grid.getNumberOfElements() );
    computeLocalErrorEstimates(errEstimates);

    double refineAbove = errEstimates.norm_linf() * _gamma;
    for (Grid::TriangleIterator iter(_grid); iter.notAtEnd(); ++iter)
        if (errEstimates[iter.getIndex()] > refineAbove)
            iter->markForRefinement();
    return errEstimates.norm_l2();
}

double ErrorEstimator::getLocalErrorEstimates(Vector & errEstimates) const
{
    computeLocalErrorEstimates(errEstimates);
    // return integrated errorEstimator
    Vector area ( _grid.getNumberOfElements() );
    for (Grid::TriangleConstIterator iter(_grid); iter.notAtEnd(); ++iter)
        area[iter.getIndex()] = iter->getVolume();
    return area * errEstimates;
}

/**********************************************************************************************/
/*                                   class DiffusiveErrorEstimator                            */
/**********************************************************************************************/
DiffusiveErrorEstimator::DiffusiveErrorEstimator(Grid & grid, TensorOrder0 & tensor, const Vector & f, const Vector & u, double gamma )
    : ErrorEstimator( grid, f, u, gamma )
    , _tensor ( tensor )
{}

DiffusiveErrorEstimator::~DiffusiveErrorEstimator()
{}

void DiffusiveErrorEstimator::computeLocalErrorEstimates(Vector & errEstimates) const
{
    for (Grid::TriangleConstIterator iter(_grid); iter.notAtEnd(); ++iter)
        errEstimates[iter.getIndex()] = diffusivityJump(*iter);
}

double DiffusiveErrorEstimator::diffusivityJump(const Triangle & t) const
{
    BarCoord center ( 1./3., 1./3., 1./3. );
    SmallVector2<> midpoint;
    SmallVector2<> otherMidpoint;
    t.calcCartesianCoord ( center, midpoint );
    double diffusivity = _tensor.evaluate( t, center );
    double otherDiffusivity;
    double ret = 0;

    // loop over all edges
    for (int e = 0; e < 3; ++e)
    {
        if (t.getNeighbourIndex(e) != IndexNotSet)
        {
            const Triangle & t_neighb = t.getNeighbour(e);

            t_neighb.calcCartesianCoord ( center, otherMidpoint );
            otherDiffusivity = _tensor.evaluate( t_neighb, center );

            ret += t.getEdgeLength(e) * abs ( diffusivity - otherDiffusivity ) / (otherMidpoint-=midpoint).norm_l2();
        }
    }
    return ret;
}


/**********************************************************************************************/
/*                                   class AposterioriErrorEstimator                          */
/**********************************************************************************************/
AposterioriErrorEstimator::AposterioriErrorEstimator(Grid & grid, TensorOrder0 & tensor, const Vector & f, const Vector & u, double gamma )
    : ErrorEstimator( grid, f, u, gamma )
    , _tensor ( tensor )
{}

AposterioriErrorEstimator::~AposterioriErrorEstimator()
{}

void AposterioriErrorEstimator::computeLocalErrorEstimates(Vector & errEstimates) const
{
    for (Grid::TriangleConstIterator iter(_grid); iter.notAtEnd(); ++iter)
        errEstimates[iter.getIndex()] = computeElementResidual(*iter) + computeSingularResidual(*iter);
}

double AposterioriErrorEstimator::computeElementResidual(const Triangle & t) const{
  SmallVector2<> gradTensor, gradU;
  t.calcCartesianGrad ( this->_u, gradU );
  double res = 0.;
  // run over all quadrature points 
  for( int i = 0; i < ElementQuadType::_numPoints; i++ )
  {
    _tensor.evaluateGrad( t, ElementQuadType::_barCoord[i], gradTensor );
    double arg = t.getLongestEdgeLength() * ( t.calcFunctionValue( this->_f, ElementQuadType::_barCoord[i] ) + gradTensor * gradU );
    res += arg * arg;
  }
  return t.getVolume() * res;
}

double AposterioriErrorEstimator::computeSingularResidual(const Triangle & t) const{
  double res = 0.;
  // run over three edges
  for (int e = 0; e < 3; ++e)
    {
        // boundary edge?
        if (t.getNeighbourIndex(e) != IndexNotSet)
        {
            const Triangle & t_neighb = t.getNeighbour(e);
            SmallVector2<> normal = t.getNormal(e);

            for( int i = 0; i < ElementQuadType::_numPoints; i++ )
            {
              SmallVector2<> gradUPlus, gradUMinus;

              t.calcCartesianGrad ( this->_u, gradUPlus );
              gradUPlus *= _tensor.evaluate( t, ElementQuadType::_barCoord[i] );
              // assume that tensor is continuous, i.e. it does not matter on which element it is is evaluated
              t_neighb.calcCartesianGrad ( this->_u, gradUMinus );
              gradUMinus *= _tensor.evaluate( t, ElementQuadType::_barCoord[i] );
              gradUMinus -= gradUPlus;
              // edge length is squared du o integration
              gradUMinus *= t.getEdgeLength(e); 
              double arg = gradUMinus * normal;
              res +=  arg * arg;
            }
        }
    }
  return res;
}


/**********************************************************************************************/
/*                                   class DoubleWellErrorIndicator                           */
/**********************************************************************************************/
DoubleWellErrorIndicator::DoubleWellErrorIndicator(Grid & grid, const Vector & f, const Vector & u, double gamma )
    : ErrorEstimator( grid, f, u, gamma )
{}

DoubleWellErrorIndicator::~DoubleWellErrorIndicator()
{}

void DoubleWellErrorIndicator::computeLocalErrorEstimates(Vector & errEstimates) const
{
    for (Grid::TriangleConstIterator iter(_grid); iter.notAtEnd(); ++iter)
        errEstimates[iter.getIndex()] = computeDoubleWell(*iter);
}

double DoubleWellErrorIndicator::computeDoubleWell(const Triangle & t) const{
    double res = 0.;
    // run over all quadrature points 
    for( int k = 0; k < ElementQuadType::_numPoints; k++ )
    {
        // Interpolate linear phasefield at quadrature point
        double local = 0.;
        for ( int i = 0; i < 3; i++ )
        {
            local += _u[ t.getNodeIndex(i) ] * ElementQuadType::_barCoord[k][i];
        }
        local *= local;
        local -= 1.;
        res += local * local * ElementQuadType::_weight[k];
    }
    return t.getVolume() * res;
}
