/** Error estimator
 *
 *  \file   errorEstimator.h
 *  \author Benedict Geihe, Behrend Heeren
 *  \date   19.10.2012
 *
 */

#ifndef __errorEstimator__h
#define __errorEstimator__h

#include "../vecmat/vector.h"
#include "../vecmat/tensor.h"
#include "../triang/grid.h"
#include "quadrature.h"

/**
 *  \brief Class for local error estimation
 *
 *  marks those triangles which contribute most to the chosen error indicator
 *
 */
class ErrorEstimator
{
    public:
        ErrorEstimator(Grid & grid, const Vector & f, const Vector & u, double gamma = 0.5);
        virtual ~ErrorEstimator();

        double markTriangles();
        double getLocalErrorEstimates(Vector & errEstimates) const;
        
    protected:
        virtual void computeLocalErrorEstimates(Vector & errEstimates) const = 0;

        Grid & _grid;
        const Vector & _f;
        const Vector & _u;
        double _gamma;
};


/**
 *  \brief Error estimatior for diffusion problems
 *
 *  an error indicator is computed based on steep transitions in the diffusivity
 *
 */
class DiffusiveErrorEstimator : public ErrorEstimator
{
    public:
        DiffusiveErrorEstimator(Grid & grid, TensorOrder0 & _tensor, const Vector & f, const Vector & u, double gamma = 0.5);
        virtual ~DiffusiveErrorEstimator();

    protected:
        virtual void computeLocalErrorEstimates(Vector & errEstimates) const;
        double diffusivityJump(const Triangle & t) const;

    TensorOrder0 & _tensor;
};


/**
 *  \brief Error estimatior for diffusion problems
 *
 *  an error indicator \$ \eta_T \$ 
 *
 */
class AposterioriErrorEstimator : public ErrorEstimator
{
    typedef BarQuad1  ElementQuadType;
    typedef LineQuad1 EdgeQuadType;
      
    public:
        AposterioriErrorEstimator(Grid & grid, TensorOrder0 & _tensor, const Vector & f, const Vector & u, double gamma = 0.5);
        virtual ~AposterioriErrorEstimator();

    protected:
        virtual void computeLocalErrorEstimates(Vector & errEstimates) const;
        double computeElementResidual(const Triangle & t) const;
        double computeSingularResidual(const Triangle & t) const;

    TensorOrder0 & _tensor;
};


/**
 *  \brief Error indicator for phase fields
 */
class DoubleWellErrorIndicator : public ErrorEstimator
{
    typedef BarQuad3 ElementQuadType;

    public:
        DoubleWellErrorIndicator(Grid & grid, const Vector & f, const Vector & u, double gamma = 0.5);
        virtual ~DoubleWellErrorIndicator();

    protected:
        virtual void computeLocalErrorEstimates(Vector & errEstimates) const;
        double computeDoubleWell(const Triangle & t) const;
};

#endif
