/** Bilinear form for two scale problems
 *
 *  \file   twoscaleBilf.cpp
 *  \author Benedict Geihe, Behrend Heeren, Sascha Tölkes
 *  \date   24.10.2012
 *
 */

#include "twoscaleBilf.h"


/**********************************************************************************************/
/*                                   class CellProblem                                        */
/**********************************************************************************************/
//! \brief Constructor
template <typename MicroBilfType>
CellProblem<MicroBilfType>::CellProblem( const MicroBilfType & bilf )
:   _grid()          // construct grid by default constructor, further processing below
,   _bilf( bilf )    // need the copy constructor here
,   _stiffMat()
,   _baseSolution1()
,   _baseSolution2()
{
    // now create the real grid
    //! \bug Generalize the initial grid size of the micro grid!
    const int gridDepth = 6;  // this is arbitrary here!
    _grid = Grid::createRectangularGrid(0., 0., 1., 1.,gridDepth );

    // solve periodic cell problem for base solutions
    recomputeBaseSolutions();
}

template <typename MicroBilfType>
void CellProblem<MicroBilfType>::recomputeBaseSolutions()
{
    _baseSolution1.resize ( _grid.getNumberOfNodes() );
    _baseSolution2.resize ( _grid.getNumberOfNodes() );

    _baseSolution1.setZero();
    _baseSolution2.setZero();

    PeriodicBoundaryMask pMask ( _grid );

    LinearFEOperator< MicroBilfType, BarQuad1 > stiffOp ( _grid, _bilf );

    // make copy of uncollapsed matrix
    _stiffMat = stiffOp;

    MassBilf mbilf;
    LinearFEOperator< MassBilf, BarQuad2 > massOp ( _grid, mbilf );

    Vector allone ( _grid );
    Vector constr ( _grid );
    allone.setAll( 1. );
    massOp.apply(allone, constr);

    // transform to periodic setting
    pMask.apply(massOp);
    pMask.collapse(constr);

    // projectingCGapply requires a normalized normal vector
    constr.normalize();

    Vector u_H_1(_grid), u_H_2(_grid), rhs_1(_grid), rhs_2(_grid);

    // generate linear profiles
    generateLinearProfile( SmallVector2<>(1.,0.), u_H_1 );
    generateLinearProfile( SmallVector2<>(0.,1.), u_H_2 );

    // use (uncollapsed) stiffness matrix to evaluate right hand side for given u_H
    stiffOp.apply( u_H_1, rhs_1 );
    stiffOp.apply( u_H_2, rhs_2 );

    // collapse stiffOp
    pMask.apply ( stiffOp );

    // bring rhs to right hand side
    rhs_1*=-1.;
    rhs_2*=-1.;

    // collapse right hand side to account for periodic test functions used
    pMask.collapse( rhs_1 );
    pMask.collapse( rhs_2 );

    // solve the obtained linear system of equations
    Solver<SparseMatrix> solver1 ( stiffOp, false );
    solver1.setMaxIter (10000);
    solver1.setTol(1E-10);
    solver1.projectingCGapply( constr, rhs_1, _baseSolution1 );
    Solver<SparseMatrix> solver2 ( stiffOp, false );
    solver2.setMaxIter (10000);
    solver2.setTol(1E-10);
    solver2.projectingCGapply( constr, rhs_2, _baseSolution2 );

    // undo periodic identification
    pMask.extend( _baseSolution1 );
    pMask.extend( _baseSolution2 );

    // construct full (corrected) profile
    _baseSolution1 += u_H_1;
    _baseSolution2 += u_H_2;
}

//! get reference to bilinear form
//!
//! \note this allow external methods to modify the internal state of this bilinear form
template <typename MicroBilfType>
MicroBilfType & CellProblem<MicroBilfType>::getBilfReference ()
{
    return _bilf;
}

//! get const reference to bilinear form
template <typename MicroBilfType>
const MicroBilfType & CellProblem<MicroBilfType>::getBilfReference () const
{
    return _bilf;
}

//! Evaluate microscopic energy
//! use linear combination of base solutions
template <typename MicroBilfType>
double CellProblem<MicroBilfType>::evaluateMicroEnergy( const SmallVector2<> & grad1, const SmallVector2<> & grad2 ) const
{
    Vector u_1(_grid), u_2(_grid), tmp(_grid);

    u_1.addMultiple( grad1[0], _baseSolution1 );
    u_1.addMultiple( grad1[1], _baseSolution2 );
    u_2.addMultiple( grad2[0], _baseSolution1 );
    u_2.addMultiple( grad2[1], _baseSolution2 );

    // return energy (misuse rhs_1 here)
    _stiffMat.apply( u_1, tmp );
    return tmp * u_2;
}

//! Returns a reference to base solution b
template <typename MicroBilfType>
Vector& CellProblem<MicroBilfType>::getBaseSolutionReference(unsigned int b)
{
  if(b == 1)
    return _baseSolution1;
  else if(b == 2)
    return _baseSolution2;
  else
  {
    throw "Requested invalid base solution!\n";
    return _baseSolution1;
  }
}

//! protected helper method to compute a linear profile
template <typename MicroBilfType>
void CellProblem<MicroBilfType>::generateLinearProfile( const SmallVector2<> & grad, Vector & linearProfile ) const
{
  for (Grid::NodeConstIterator nIter = _grid; nIter.notAtEnd(); ++nIter)
    linearProfile[ nIter.getIndex() ] = grad * (*nIter);
}



/**********************************************************************************************/
/*                                   class TwoscaleBilf                                       */
/**********************************************************************************************/
//! \brief Constructor
//! Upon construction provide everything needed to solve a cell problem in each macroscopic triangle
//! Start with a representative CellProblem copied to every cell
//!
//! \note Currently we assume that there is only one quadrature point per triangle!
//! \bug Generalize the number of microcells (currently we assume only one micro cell per triangle!)
template <typename MicroBilfType>
TwoscaleBilf<MicroBilfType>::TwoscaleBilf( const Grid & macroGrid, const MicroBilfType & microBilfRepresentative )
    : _macroGrid( macroGrid )
    , _cellProblems( macroGrid.getNumberOfElements(), CellProblem<MicroBilfType>( microBilfRepresentative ) )
    , _cellProblemChanged ( macroGrid.getNumberOfElements() )
{ }

//! get reference to microscopic bilinear forms
//!
//! \note this allow external methods to modify the internal state of microscopic bilinear forms
template <typename MicroBilfType>
MicroBilfType & TwoscaleBilf<MicroBilfType>::getBilfReference ( unsigned int index )
{
    assert ( index < _cellProblems.size() );
    // we assume that getting a non const reference implies a modification
    _cellProblemChanged[index] = true;
    return _cellProblems[index].getBilfReference();
}

//! get const reference to microscopic bilinear forms
template <typename MicroBilfType>
const MicroBilfType & TwoscaleBilf<MicroBilfType>::getBilfReference ( unsigned int index ) const
{
    assert ( index < _cellProblems.size() );
    return _cellProblems[index].getBilfReference();
}

//! update cell problems
template <typename MicroBilfType>
void TwoscaleBilf<MicroBilfType>::update()
{
    //! \bug we still need an intelligent initialization here!
    if ( _macroGrid.getNumberOfElements() != _cellProblems.size() )
    {
        // only relevant if macroGrid was refined
        //! \bug make copies in an intelligent way
        _cellProblems.reserve( _macroGrid.getNumberOfElements() );
        _cellProblemChanged.reserve( _macroGrid.getNumberOfElements() );
    }

    //! \bug do parallelization here!
    for ( unsigned int i = 0; i < _cellProblemChanged.size(); ++i )
        if ( _cellProblemChanged[i] )
        {
            _cellProblems[i].recomputeBaseSolutions();
            _cellProblemChanged[i] = false;
        }
}

//! Evaluate macroscopic bilinear form by computing microscopic corrections
template <typename MicroBilfType>
double TwoscaleBilf<MicroBilfType>::operator() (const Triangle & t, const BarCoord & /*q*/, int i, int j ) const
{
  // obtain cartesian gradients for base functions \phi_i and \phi_j
  // as we use linear elements only the gradients are constant, higher order quadrature would be no use
  SmallVector2<> cartGradi, cartGradj;
  t.calcCartesianBaseGrad ( i, cartGradi );
  t.calcCartesianBaseGrad ( j, cartGradj );

  return _cellProblems[t.getIndex()].evaluateMicroEnergy( cartGradi, cartGradj );
}


// Instantiation
template class CellProblem<IsoDiffusiveStiffBilf>;
template class TwoscaleBilf<IsoDiffusiveStiffBilf>;
