/** Lab exercise 3
 *
 *  \file   cellAdaptive.cpp
 *  \author Benedict Geihe, Behrend Heeren, Sascha Toelkes
 *  \date   31.10.2012
 *
 * --------------------------------------------------------------------------------------------- */


#include <iostream>

using namespace std;

#include "fe/LinearFEOperator.h"
#include "fe/diffusiveBilf.h"
#include "fe/boundaryMasks.h"
#include "fe/errorEstimator.h"
#include "vecmat/solver.h"

//! evaluates a linear function increasing in direction normal to n with slope m
double linearProfile (const Node & x)
{
    SmallVector2<> n (1., 2.);
    const double m  = 1.;

    return m / n.norm_l2q() * ( n[1]*x[0] - n[0]*x[1] );
}

//! main
int main()
{
  try{
    char filename[25];

    // Generate a triangulation of the unit square
    int gridDepth = 4;
    // special grid class to assure feasible refinements in a periodic setting
    PeriodicSquare grid( gridDepth );

    int iteration = 0;
    while ( iteration < 10 )
    {
        ++ iteration;

        // Create vectors for coefficients of solution u_h, given profile u_H, right hand side and constraint for projecting CG
        Vector u_h(grid),
              u_H(grid),
              rhs(grid),
              constr(grid);

        // Set u_H using function given above
        for (Grid::NodeConstIterator nIter = grid; nIter.notAtEnd(); ++nIter)
          u_H[ nIter.getIndex() ] = linearProfile(*nIter);

        // Plot u_H
        grid.plotSolution(u_H, "u_H", JPG, true);

        // Create the diffusion tensor needed for evaluation of the bilinear form
        BlobTensor tensor ( SmallVector2<>(.2,.2), .1, .5 );

        // Create the bilinear form
        IsoDiffusiveStiffBilf bilf( tensor );

        // Create the linear finite element operator
        LinearFEOperator< IsoDiffusiveStiffBilf, BarQuad1 > stiffOp(grid,bilf);

        // Create boundary mask to handle periodic indentification of nodes
        PeriodicBoundaryMask pMask ( grid );

        // use (uncollapsed) stiffness matrix to evaluate right hand side for given u_H
        stiffOp.apply( u_H, rhs );

        // bring rhs to right hand side
        rhs*=-1.;

        // collapse right hand side to account for periodic test functions used
        pMask.collapse( rhs );

        // plot right hand side
        sprintf(filename, "rhs_%i", iteration);
        grid.plotSolution(rhs, filename, JPG, true);

        // now collapse stiffness matrix used on left hand side to account for periodicity
        pMask.apply( stiffOp );

        // Create a standard massMatrix
        MassBilf mbilf;
        LinearFEOperator< MassBilf, BarQuad2 > massOp(grid,mbilf);

        // Create a vector with all entries 1 to compute integrals of u_h:  u_h^t M 1
        Vector allone ( constr.size() );
        allone.setAll( 1. );
        massOp.apply( allone, constr );

        // transform to periodic setting
        pMask.collapse( constr );

        // projectingCGapply requires a normalized normal vector
        constr.normalize();

        // solve the obtained linear system of equations
        Solver<SparseMatrix> solver(stiffOp, false);
        solver.setMaxIter (10000);
        solver.setTol(1E-10);
        solver.projectingCGapply( constr, rhs, u_h);

        // to get a nicer output assign values to collapsed boundary nodes
        pMask.extend( u_h );

        // plot the periodic profile
        sprintf(filename, "u_h_%i", iteration);
        grid.plotSolution(u_h, filename, JPG, true);

        // plot the full profile
        u_H += u_h;
        sprintf(filename, "correctedProfile_%i", iteration);
        grid.plotSolution(u_H, filename, JPG, true);

        // do refinement
        DiffusiveErrorEstimator errorEstimator( grid, tensor, rhs, u_h, .8 );
        errorEstimator.markTriangles();
        sprintf(filename, "domain_%i", iteration);
        grid.plot(filename, JPG, true, true);

        grid.refineMarkedTriangles();
        cerr << "New number of nodes: " << grid.getNumberOfNodes() << endl;
    }
  }
  catch ( string &err ) { cerr << err << endl; }

  return 0;
}
