/** Test for two scale code
 *
 *  \file   twoscale.cpp
 *  \author Benedict Geihe, Behrend Heeren, Sascha Toelkes
 *  \date   21.11.2012
 *
 *  Solve the (two-scale) problem
 *
 *    -div( a(x,x/\epsilon) \nabla u^\epsilon ) = f in \Omega
 *                                            u = 0 on \partial\Omega
 *
 *  with Omega = [0,1]^2 and compare to the single scale version.
 *
 *
 * --------------------------------------------------------------------------------------------- */


#include <iostream>
#include <stdlib.h>

using namespace std;

#include "fe/LinearFEOperator.h"
#include "fe/bilf.h"
#include "fe/diffusiveBilf.h"
#include "fe/twoscaleBilf.h"
#include "fe/boundaryMasks.h"
#include "vecmat/solver.h"
#include "vecmat/tensor.h"


//!  \todo Complete the class \a TwoscaleBilf in fe/twoscaleBilf.{h,cpp} 
//!  \todo Fill the main() according to the instructions below, i.e. compare single scale to two scale problem!
 
int main()
{
    // construct macro grid
    int gridDepth = 8;
    Grid macroGrid = Grid::createRectangularGrid( 0., 0., 1., 1., gridDepth );

    DirichletBoundaryMask bMask ( macroGrid );

    Vector u(macroGrid),
           u_ts(macroGrid),
           rhs(macroGrid),
           f(macroGrid);

    // source term
    for (Grid::NodeConstIterator nIter = macroGrid; nIter.notAtEnd(); ++nIter)
      f[ nIter.getIndex() ] = (SmallVector2<> ( .5, .5 ) -= (*nIter)).norm_l2();
    char sol_filename[12];
    sprintf(sol_filename, "f");
    macroGrid.plotSolution(f, sol_filename, JPG, true);

    // prepare right hand side
    MassBilf mBilf;
    LinearFEOperator< MassBilf, BarQuad2 > massOp(macroGrid,mBilf);
    massOp.apply( f, rhs );
    sprintf(sol_filename, "rhs");
    macroGrid.plotSolution(rhs, sol_filename, JPG, true);

    // usual stiffness operator
    StiffBilf sBilf;
    LinearFEOperator< StiffBilf, BarQuad1 > stiffOp(macroGrid,sBilf);

    // two-scale stiffness operator
    BlobTensor tensor ( SmallVector2<>(.5,.5), .05 );
    IsoDiffusiveStiffBilf bilfRepresentative ( tensor );
    TwoscaleBilf<IsoDiffusiveStiffBilf> tsBilf( macroGrid, bilfRepresentative );
    // now forget about tensor and bilfRepresentative, they will be copied internally
    // for bilfRepresentative two base solution will be computed and copied everywhere else

    // we still have control over tsBilf: manipulate the microscale problems by accessing their bilinear form
    for (Grid::TriangleConstIterator tIter = macroGrid; tIter.notAtEnd(); ++tIter)
    {
        IsoDiffusiveStiffBilf & targetBilf = tsBilf.getBilfReference (tIter->getIndex());
        BlobTensor * tensorPtr = dynamic_cast<BlobTensor*>( targetBilf.getTensorPtr() );
        tensorPtr->setInflection ( .4 * tIter->getCenter().norm_l2() );
    }
    // as the microscale problems have changed now call update on twoscaleBilf
    tsBilf.update();
    // now do the assembly
    LinearFEOperator< TwoscaleBilf<IsoDiffusiveStiffBilf>, BarQuad1 > tsstiffOp(macroGrid,tsBilf);

    // set Dirichlet boundary
    bMask.apply( stiffOp );
    bMask.apply( tsstiffOp );
    bMask.apply( rhs );

    // solve this system using single scale and two-scale stiffness operator
    Solver<SparseMatrix> solver(stiffOp, true);
    solver.setMaxIter (1000);
    solver.setTol(1E-8);
    solver.CGapply(rhs, u);
    Solver<SparseMatrix> tssolver(tsstiffOp, true);
    tssolver.setMaxIter (1000);
    tssolver.setTol(1E-8);
    tssolver.CGapply(rhs, u_ts);

    sprintf(sol_filename, "solution");
    macroGrid.plotSolution(u, sol_filename, JPG, true);
    sprintf(sol_filename, "solution_ts");
    macroGrid.plotSolution(u_ts, sol_filename, JPG, true);

#if defined(__MINGW32_VERSION) || defined(_MSC_VER) || defined(__BCPLUSPLUS__)
    system("PAUSE");
#endif
    return 0;
}
