/** Lab exercise 6
 * 
 *  \file   problem06.cpp
 *  \author Benedict Geihe, Behrend Heeren
 *  \date   12.12.2012
 *
 *  - Use the provided routines to create phasefields for a circle and a bar
 *  - Compare your approximation to the known true values
 *  - Try to decrease the error as much as possible by:
 *    - Changing epsilon and the global mesh size h. What can be said about their relation?
 *    - Doing local adaptive mesh refinements. For this purpose you need to implement an error
 *      indicator like on problem sheet 2. Think about where the crucial regions of the
 *      phasefields are and how they could be detected by an error indicator.
 *
 * --------------------------------------------------------------------------------------------- */



#include <iostream>
#include <stdlib.h>

using namespace std;

#include "fe/bilf.h"
#include "fe/LinearFEOperator.h"
#include "fe/PhasefieldOps.h"
#include "fe/NonlinearFEOperator.h"
#include "fe/errorEstimator.h"


void generateCircularPhasefield ( Vector & pf, const Grid & grid, const SmallVector2<> & mid, const double radius, const double eps )
{
  for (Grid::NodeConstIterator nIter = grid; nIter.notAtEnd(); ++nIter)
  {
    SmallVector2<> dist ( mid );
    dist -= (*nIter);
    pf[ nIter.getIndex() ] = tanh ( 3./4. * ( radius - dist.norm_l2() ) / eps );
  }
}

void generateBarPhasefield ( Vector & pf, const Grid & grid, const double width, const double y, const double eps )
{
  for (Grid::NodeConstIterator nIter = grid; nIter.notAtEnd(); ++nIter)
  {
    pf[ nIter.getIndex() ] = tanh ( 3./4. * ( width - fabs( (*nIter)[1]-y ) ) / eps );
  }
}


// #define CIRCLE

int main ( int argc, char * argv[] )
{
    char filename[20];
    int    gridDepth   = 10;
    int    refinements = 0;
    double eps         = 0.04;
    double real_vol, real_length;

    // get parameters
    if ( argc > 1 ) eps         = atof ( argv[1] );
    if ( argc > 2 ) gridDepth   = atoi ( argv[2] );
    if ( argc > 3 ) refinements = atoi ( argv[3] );

    Grid grid = Grid::createRectangularGrid( 0., 0., 1., 1., gridDepth );

    Vector pf(grid);

    double h = 1.;
    for ( int i = 0; i < gridDepth; ++i ) h /= M_SQRT2;
    cout << endl;
    cout << "Epsilon:            " << eps << endl;
    cout << "Mesh size:          " << h << endl;
    cout << "Local refinements:  " << refinements << endl;

#ifdef CIRCLE
    SmallVector2<> midpoint (.5,.5);
    double circ_radius = .3;
    real_vol = circ_radius * circ_radius * M_PI;
    real_length = 2 * circ_radius * M_PI;
    generateCircularPhasefield( pf, grid, midpoint, circ_radius, eps );
#else
    double bar_width   = .2;
    double bar_ypos    = .6;
    real_vol = bar_width;
    real_length = 2.;
    generateBarPhasefield( pf, grid, bar_width, bar_ypos, eps );
#endif

    for ( int r = 0; r < refinements; ++r )
    {
        DoubleWellErrorIndicator( grid, pf, pf, .8 ).markTriangles();
        sprintf(filename, "domain_%i", r);
        grid.plot(filename, JPG, false, true);
        grid.refineMarkedTriangles();
        pf.resize( grid.getNumberOfNodes() );
#ifdef CIRCLE
        generateCircularPhasefield( pf, grid, midpoint, circ_radius, eps );
#else
        generateBarPhasefield( pf, grid, bar_width, bar_ypos, eps );
#endif
    }
    cout << "Number of elements: " << grid.getNumberOfElements() << endl << endl;

    sprintf( filename, "phasefield");
    grid.plotSolution(pf, filename, JPG, true);

    // quadratic parts may be treated by using linear ops
    MassBilf mBilf;
    LinearFEOperator< MassBilf, BarQuad2 > massOp(grid,mBilf);
    StiffBilf sBilf;
    LinearFEOperator< StiffBilf, BarQuad1 > stiffOp(grid,sBilf);


    // compute volume of phasefield
    double vol = 0.;
    Vector tmp(grid);
    Vector allone(grid);
    allone.setAll(1.);

    massOp.apply ( pf, tmp );
    vol = tmp * pf;
    vol += 2. * ( tmp * allone );
    massOp.apply ( allone, tmp );
    vol += tmp * allone;
    vol *= .25;

    cout << "Object has volume          \t" << vol << " \t(error " << fabs(vol-real_vol) << ")" << endl;


    // compute length of phasefield
    double length = 0.;
    // nonlinearity
    NonlinearFEOperator<double, BarQuad3, PhasefieldNonLin> pfnonlin(grid);
    pfnonlin.applyAdd( pf, length );
    length *= 9. / ( 16. * eps );
    stiffOp.apply ( pf, tmp );
    length += eps * ( tmp * pf );
    length *= .5;

    cout << "Object has perimeter length\t" << length << " \t(error " << fabs(length-real_length) << ")" << endl << endl;


#if defined(__MINGW32_VERSION) || defined(_MSC_VER) || defined(__BCPLUSPLUS__)
    system("PAUSE");
#endif
    return 0;
}
