/** Lab exercise 7
 *
 *  \file   problem07.cpp
 *  \author Benedict Geihe
 *  \date   18.12.2012
 *
 *  Complete the bilinear form \a PhasefieldElastStiffBilf needed to set up the stiffness
 *  matrix for the linearized elastic problem.
 *  Review the explicit integration of the nonzero Neumann boundary data below
 *  and enjoy a merry Christmas!
 *
 * --------------------------------------------------------------------------------------------- */



#include <cmath>
#include <iostream>

using namespace std;

#include "fe/LinearFEOperator.h"
#include "fe/elastBilf.h"
#include "fe/boundaryMasks.h"
#include "vecmat/solver.h"
#include "mystery.h"



void prepareBoundaryData ( const Grid & grid, const SmallVector2<> & load, Vector & rhs, DirichletBoundaryMask & bdryMask )
{
    const int N = grid.getNumberOfNodes();
    for ( Grid::BoundaryConstIterator bIter = grid; bIter.notAtEnd(); ++bIter )
    {
        // Determine region for nonzero Neumann boundary data
        if (  ( ( fabs( bIter.getStart()[1] > 5.5 ) ) || ( fabs( bIter.getEnd()[1] > 5.5 ) ) )
             && ( fabs( bIter.getStart()[0] < 1e-14 ) ) && ( fabs( bIter.getEnd()[0] < 1e-14 ) ) )
        {
            // Integration on one segment
            // There are two linear finite element hat functions on this segment
            // Each one is multiplied by the constant load vector
            rhs[     bIter.getStartIndex() ] += bIter.getLength() * .5 * load[0];
            rhs[     bIter.getEndIndex()   ] += bIter.getLength() * .5 * load[0];
            rhs[ N + bIter.getStartIndex() ] += bIter.getLength() * .5 * load[1];
            rhs[ N + bIter.getEndIndex()   ] += bIter.getLength() * .5 * load[1];
        }

        // Homogeneous Dirichlet boundary data at bottom
        if ( fabs( bIter.getStart()[1] ) < 1e-14 )
        {
            bdryMask.mask(     bIter.getStartIndex() );
            bdryMask.mask( N + bIter.getStartIndex() );
        }
    }
}


int main()
{
  try  {
    // generate the grid
    Grid grid;
    generateMysteriousGrid( grid );
    const int N = grid.getNumberOfNodes();
    grid.plot ( "domain.eps", JPG, true );

    // create a phasefield
    Vector pf(grid);
    generateMysteriousPhasefield( grid, pf );
    grid.plotSolution(pf, "phasefield", JPG, true);

    // surface loading
    SmallVector2<> load ( .001, -.001 );

    // prepare boundary data
    Vector rhs(2*N);
    DirichletBoundaryMask bdryMask;
    //! \todo review the explicit integration of the nonzero Neumann boundary data
    prepareBoundaryData( grid, load, rhs, bdryMask );

    // create the bilinear form
    PhasefieldElastStiffBilf bilf ( pf );
    // create the vector valued linear finite element operator
    LinearFEOperator<PhasefieldElastStiffBilf, BarQuad1, 2> stiffOp(grid, bilf);

    // apply the boundary mask
    bdryMask.apply( stiffOp );

    // solve
    Vector u(2*N);
    Solver<SparseMatrix> solver(stiffOp, true);
    solver.setMaxIter (10000);
    solver.CGapply(rhs, u);

    grid.plotSolutionDeformed(pf,u,"solution",JPG );

  } catch ( string err ) { cerr << err << endl; }
  return 0;
}
