/** Klasse zur Verwaltung einer Triangulierung
 *
 *  \file   grid.cpp
 *  \author Stefan W. von Deylen, Benedict Geihe
 *  \date   18.09.2008
 *
 */

#include "grid.h"

#include <sstream>
#include <cassert>
#include <cstdlib>

#if defined(__MINGW32_VERSION) || defined(_MSC_VER) || defined(__BCPLUSPLUS__)
#include <process.h>
#define GNUPLOT_PATH string("\"C:\\Program Files\\gnuplot\\bin\\wgnuplot.exe\" ")
#else
#include <unistd.h>
#define GNUPLOT_PATH string("gnuplot ")
#endif


using namespace std;

//! Standard-Konstruktor
Grid::Grid()
    : _nodeList()
    , _triangleList()
{}
// ------------------------------------------------------------------------------------------------

//! \brief Konstruktor: Erstellt ein Gitter aus 3 Knoten
//! Nach Aufruf enthaelt das Gitter 3 Knoten mit Indices 0,1,2 und ein Dreieck (0,1,2)
Grid::Grid( const Node & node1, const Node & node2, const Node & node3 )
{
    _nodeList.push_back(node1);
    _nodeList.push_back(node2);
    _nodeList.push_back(node3);

    // erstelle erstes Dreieck
    _triangleList.push_back( Triangle(0, 1, 2) );
    // setze Index des Dreiecks und "sein" Gitter
    Triangle & newTriangle = _triangleList[0];
    newTriangle.setIndex(0);
    newTriangle.setGrid(this);
    newTriangle.computeTransformation();
}
// ------------------------------------------------------------------------------------------------

//! Kopier-Konstruktor
Grid::Grid( const Grid & otherGrid )
    : _nodeList( otherGrid._nodeList )
    , _triangleList( otherGrid._triangleList )
{
    // setze Gitter-Zeiger der Dreiecke
    for ( TriangleIterator tIter = *this; tIter.notAtEnd(); ++tIter )
    {
        tIter->setGrid(this);
    }
}
// ------------------------------------------------------------------------------------------------

//! Zuweisungs-Operator
Grid & Grid::operator=( const Grid & otherGrid )
{
    if ( this == &otherGrid ) return *this;

    // kopiere die Daten
    _nodeList =     otherGrid._nodeList;
    _triangleList = otherGrid._triangleList;

    // setze Gitter-Zeiger der Dreiecke
    for ( TriangleIterator tIter = *this; tIter.notAtEnd(); ++tIter )
    {
        tIter->setGrid(this);
    }

    return *this;
}
// ------------------------------------------------------------------------------------------------

//! Fuegt einen einzelnen Knoten zum Gitter hinzu
//! \return Globaler Index des neuen Knotens
GlobalIndex Grid::addNode( const Node & nodeToAdd )
{
    _nodeList.push_back( nodeToAdd );
    return _nodeList.size() - 1;
}
// ------------------------------------------------------------------------------------------------

//! \brief  Fuegt einen neuen Knoten und ein neues Dreieck an einer existierenden Seite hinzu
//! \param  nodeToAdd    Knoten (bereits konstruiert) der hinzugefuegt werden soll
//! \param  existentEdge Dart, der die Seite eines Dreiecks angibt, an der das neue Dreieck anliegen soll
//! \return              Globaler Index des neuen Dreiecks
//! \warning             An der angegebenen Seite darf noch kein Dreieck liegen!
GlobalIndex Grid::addNodeAndTriangle(const Node & nodeToAdd, DartIterator existentEdge)
{
    // assert that neighbTriangle has no neighbour on side createOnSide yet.
    assertEdgeIsBoundary(existentEdge);

    // add given node to node list
    GlobalIndex newNodeIndex = addNode(nodeToAdd);
    GlobalIndex node1 = existentEdge.getGlobalNodeIndex();
    GlobalIndex node2 = existentEdge.getNextNodeGlobalIndex();

    // create new triangle
    Triangle & newTriangle = getTriangle(addTriangle(newNodeIndex, node1, node2));

    // set neighbour relations
    newTriangle.setNeighbourIndex(0, existentEdge.getGlobalTriangleIndex());
    existentEdge.getTriangle().setNeighbourIndex(existentEdge.getLocalEdgeIndex(), newTriangle.getIndex());

    return newTriangle.getIndex();
}
// ------------------------------------------------------------------------------------------------

//! \brief  Fuegt ein neues Dreieck an 2 existierenden Seiten hinzu
//! \param  existentEdge1 Dart, der die Seite des 1. Dreiecks angibt, an der das neue Dreieck anliegen soll
//! \param  existentEdge2 Dart, der die Seite des 2. Dreiecks angibt, an der das neue Dreieck anliegen soll
//! \return               Globaler Index des neuen Dreiecks
//! \warning              An den angegebenen Seiten duerfen noch kein Dreiecke liegen!<br>
//!                       Die angegebenen Seiten muessen einen Knoten gemeinsam haben!
GlobalIndex Grid::addTriangle( DartIterator existentEdge1, DartIterator existentEdge2 )
{
    // assert that both triangles have one edge in common
    GlobalIndex commonNode = existentEdge1.getCommonNodeGlobalIndex(existentEdge2);

    // Find the other respective other ends of the dars
    GlobalIndex node1 = (commonNode == existentEdge1.getGlobalNodeIndex()
                        ? existentEdge1.getNextNodeGlobalIndex()
                        : existentEdge1.getGlobalNodeIndex() );
    GlobalIndex node2 = (commonNode == existentEdge2.getGlobalNodeIndex()
                        ? existentEdge2.getNextNodeGlobalIndex()
                        : existentEdge2.getGlobalNodeIndex() );

    assert ( node1 != node2 );

    // assert that the given neighbourhoods of triang 1 and 2
    // are still empty
    assertEdgeIsBoundary(existentEdge1);
    assertEdgeIsBoundary(existentEdge2);

    // create new triangle
    Triangle & newTriangle = getTriangle( addTriangle( commonNode, node1, node2 ) );

    // set neighbour relations
    newTriangle.setNeighbourIndex(2, existentEdge1.getGlobalTriangleIndex());
    newTriangle.setNeighbourIndex(1, existentEdge2.getGlobalTriangleIndex());

    existentEdge1.getTriangle().setNeighbourIndex(existentEdge1.getLocalEdgeIndex(), newTriangle.getIndex());
    existentEdge2.getTriangle().setNeighbourIndex(existentEdge2.getLocalEdgeIndex(), newTriangle.getIndex());

    return newTriangle.getIndex();
}
// ------------------------------------------------------------------------------------------------

//! \brief  Fuegt ein neues Dreieck an 3 existierenden Seiten hinzu
//! \param  existentEdge1 Dart, der die Seite des 1. Dreiecks angibt, an der das neue Dreieck anliegen soll
//! \param  existentEdge2 Dart, der die Seite des 2. Dreiecks angibt, an der das neue Dreieck anliegen soll
//! \param  existentEdge3 Dart, der die Seite des 3. Dreiecks angibt, an der das neue Dreieck anliegen soll
//! \return               Globaler Index des neuen Dreiecks
//! \warning              An der angegebenen Seiten duerfen noch keine Dreieck liegen!<br>
//!                       Die angegebenen Seiten muessen paarweise einen Knoten gemeinsam haben!
GlobalIndex Grid::addTriangle( DartIterator existentEdge1, DartIterator existentEdge2, DartIterator existentEdge3 )
{
    // assert that triangles 1 and 2 have one edge in common
    GlobalIndex commonNode12 = existentEdge1.getCommonNodeGlobalIndex(existentEdge2);

    // Find the other respective other ends of the dars
    GlobalIndex node1 = (commonNode12 == existentEdge1.getGlobalNodeIndex()
                        ? existentEdge1.getNextNodeGlobalIndex()
                        : existentEdge1.getGlobalNodeIndex() );
    GlobalIndex node2 = (commonNode12 == existentEdge2.getGlobalNodeIndex()
                        ? existentEdge2.getNextNodeGlobalIndex()
                        : existentEdge2.getGlobalNodeIndex() );

    if (   !( existentEdge3.getGlobalNodeIndex() == node1 && existentEdge3.getNextNodeGlobalIndex() == node2 )
        && !( existentEdge3.getGlobalNodeIndex() == node2 && existentEdge3.getNextNodeGlobalIndex() == node1 ) )
    {
        stringstream err_msg;
        err_msg << "Tried to create a triangle with nodes "
                << existentEdge1.getNode() << " and " << existentEdge2.getNode()
                << " which should belong to edge " << existentEdge3.getLocalEdgeIndex()
                << " of triangle " << existentEdge3.getTriangle()
                << ", but they do not.";
        throw err_msg.str();
    }

    // assert that the given neighbourhoods of triang 1 and 2
    // are still empty
    assertEdgeIsBoundary(existentEdge1);
    assertEdgeIsBoundary(existentEdge2);
    assertEdgeIsBoundary(existentEdge3);

    // create new triangle
    Triangle & newTriangle = getTriangle( addTriangle( commonNode12, node1, node2 ) );

    // set neighbour relations
    newTriangle.setNeighbourIndex( 0, existentEdge3.getGlobalTriangleIndex() );
    newTriangle.setNeighbourIndex( 1, existentEdge2.getGlobalTriangleIndex() );
    newTriangle.setNeighbourIndex( 2, existentEdge1.getGlobalTriangleIndex() );

    existentEdge1.getTriangle().setNeighbourIndex(existentEdge1.getLocalEdgeIndex(), newTriangle.getIndex());
    existentEdge2.getTriangle().setNeighbourIndex(existentEdge2.getLocalEdgeIndex(), newTriangle.getIndex());
    existentEdge3.getTriangle().setNeighbourIndex(existentEdge3.getLocalEdgeIndex(), newTriangle.getIndex());

    return newTriangle.getIndex();
}
// ------------------------------------------------------------------------------------------------

//! Fuegt ein neues Dreieck mit 3 bereits konstruierten Knoten hinzu
GlobalIndex Grid::addTriangle( GlobalIndex node0, GlobalIndex node1, GlobalIndex node2 )
{
    // create new triangle and set its index
    _triangleList.push_back( Triangle( node0, node1, node2 ) );
    Triangle & newTriangle = _triangleList[_triangleList.size()-1];
    newTriangle.setIndex(_triangleList.size() - 1);
    newTriangle.setGrid(this);
    newTriangle.computeTransformation();

    return newTriangle.getIndex();
}
// ------------------------------------------------------------------------------------------------

//! Erstellt ein regulaeres Rechtecksgitter
Grid Grid::createRectangularGrid(double pt1_x, double pt1_y, double pt2_x, double pt2_y, int cnt_refine)
{
    vector<Node> nodes;
    nodes.push_back(Node(pt1_x, pt1_y)); // 0
    nodes.push_back(Node(pt2_x, pt2_y)); // 1
    nodes.push_back(Node(pt1_x, pt2_y)); // 2
    nodes.push_back(Node(pt2_x, pt1_y)); // 3

    Grid ret(nodes[0], nodes[1], nodes[2]);
    ret.addNodeAndTriangle(nodes[3], DartIterator( ret.getTriangle( 0 ), 2));

    for (int j = 0; j < cnt_refine; ++j)
        ret.refineAll();

    return ret;
}

// ------------------------------------------------------------------------------------------------

//! Zugriff auf Knoten zu gegebenem globalen Index
Node & Grid::getNode( GlobalIndex globalNodeIndex )
{
    return _nodeList[globalNodeIndex];
}
// ------------------------------------------------------------------------------------------------

//! Zugriff auf Knoten zu gegebenem globalen Index (const)
const Node & Grid::getNode(GlobalIndex globalNodeIndex) const
{
    return _nodeList[globalNodeIndex];
}
// ------------------------------------------------------------------------------------------------

//! Zugriff auf Dreieck zu gegebenem globalen Index
Triangle & Grid::getTriangle(GlobalIndex globalTriangleIndex)
{
    assert( globalTriangleIndex != IndexNotSet );
    return _triangleList[globalTriangleIndex];
}
// ------------------------------------------------------------------------------------------------

//! Zugriff auf Dreieck zu gegebenem globalen Index (const)
const Triangle & Grid::getTriangle(GlobalIndex globalTriangleIndex) const
{
    return _triangleList[globalTriangleIndex];
}
// ------------------------------------------------------------------------------------------------

//! Liefert Anzahl der Knoten im Gitter
unsigned int Grid::getNumberOfNodes() const
{
    return _nodeList.size();
}
// ------------------------------------------------------------------------------------------------

unsigned int Grid::getNumberOfElements() const
{
    return _triangleList.size();
}
// ------------------------------------------------------------------------------------------------

//! Prueft ob sich der DartIterator am Gebietsrand befindet
void Grid::assertEdgeIsBoundary( const DartIterator& edge ) const
{
    if ( edge.getNextTriangleIndex() != IndexNotSet )
    {
        stringstream err_msg;
        err_msg << edge.getTriangle() << " already has a neighbour on side "
                << edge.getLocalEdgeIndex() << ".";
        throw err_msg.str();
    }
}
// ------------------------------------------------------------------------------------------------

//! Verfeinerung: Aus jedem Dreieck werden vier kongruente
//! @warning Nachbarschaftindices werden berechnet evtl. ohne dass die Dreiecke bereits existieren
void Grid::refine4All()
{
    // Merke die Anzahl der Dreiecke vor der Verfeinerung
    int triangleCount = _triangleList.size();

    // Fuer jedes Dreieck: Kein TriangleIterator, da sich das Gitter aendern wird
    for ( int tIndex = 0; tIndex < triangleCount; tIndex++ )
    {
        Triangle & tIter = getTriangle( tIndex );
        // Sammle Daten fuer neue Knoten (Mittelpunkte), alte Knoten (Ecken) und Nachbarindices
        int newNodes[3];
        int oldNodes[3];
        int neighbours[6];
        // Fuer jede Seite
        for ( int i = 0; i < 3; i++ )
        {
            // Laufe von Knoten 0 in Richtung Knoten 1 usw.
            DartIterator d ( tIter, (2+i)%3, i );
            oldNodes[i] = d.getGlobalNodeIndex();
            // Falls diese Seite am Rand liegt kann ein neuer Knoten hinzugefuegt werden
            if ( !d.canFlipTriangle() )
            {
                 newNodes[i]       = addEdgeMidpoint( d );
                 neighbours[2*i  ] = IndexNotSet;
                 neighbours[2*i+1] = IndexNotSet;
            }
            else
            {
                // Falls das andere Dreieck noch nicht verfeinert wurde, kann ein neuer Knoten hinzugefuegt werden
                if ( d.getNextTriangleIndex() > d.getGlobalTriangleIndex() )
                {
                    // Errechne die Indices der Nachbardreiecke (evtl. noch gar nicht vorhanden!)
                    int nindex = triangleCount + 3 * d.getNextTriangleIndex();
                    newNodes[i] = addEdgeMidpoint( d );
                    // Beachte Reihenfolge in der Dreiecke hinzugefuegt werden
                    d.flipTriangle();
                    neighbours[2*i  ] = nindex + d.getLocalNodeIndex();
                    neighbours[2*i+1] = nindex + d.getNextNodeLocalIndex();
                }
                // Ansonsten existiert das andere Dreieck bereits
                else
                {
                    Triangle & other = getTriangle( d.getNextTriangleIndex() );
                    // Suche nach dem Eckpunkt der auf aktueller Kante liegt
                    for ( int j = 0; j < 3; j++ )
                    {
                        if ( tIter.getEdgeMidpoint((i+2)%3) == other.getNode(j) )
                        {
                            newNodes[i] = other.getNodeIndex(j);
                            // Bewege den Iterator ins naechste Dreieck und dort zur aktuellen Seite
                            DartIterator dOther1( other, (j+1)%3, j );
                            dOther1.flipTriangle();
                            dOther1.flipEdge();
                            dOther1.flipNode();
                            // Bewege einen zweiten in die andere Richtung
                            DartIterator dOther2( other, (j+1)%3, j );
                            dOther2.flipEdge();
                            dOther2.flipTriangle();
                            // Entweder die Eckpunkte stimmen ueberein...
                            if ( dOther1.getGlobalNodeIndex() == d.getGlobalNodeIndex() )
                            {
                                neighbours[2*i  ] = dOther1.getGlobalTriangleIndex();
                                neighbours[2*i+1] = dOther2.getGlobalTriangleIndex();
                            }
                            // ...oder nicht
                            else
                            {
                                neighbours[2*i  ] =  dOther2.getGlobalTriangleIndex();
                                neighbours[2*i+1] =  dOther1.getGlobalTriangleIndex();
                            }
                        }
                    }
                }
            }
        }
        // Fuege 3 neue Dreiecke in den Ecken hinzu und mache aus dem alten das zentrale Dreieck
        for ( int i = 0; i < 3; i++ )
        {
            // Knoten 0 ist ehemalige Ecke
            Triangle & newTriangle = getTriangle( addTriangle( oldNodes[i], newNodes[i], newNodes[(2+i)%3] ) );
            // Setze Nachbarn
            newTriangle.setNeighbourIndex( 0, tIndex );
            newTriangle.setNeighbourIndex( 1, neighbours[(2*i+5)%6] ); // das ist 2*i-1 im geeigneten Sinne
            newTriangle.setNeighbourIndex( 2, neighbours[ 2*i     ] );
            newTriangle.computeTransformation();

            Triangle & oldTriangle = getTriangle( tIndex );
            oldTriangle.setNodeIndex( i, newNodes[i] );
            oldTriangle.setNeighbourIndex( (i+1)%3, newTriangle.getIndex() );
        }
        Triangle & oldTriangle = getTriangle( tIndex );
        oldTriangle.computeTransformation();
    }
}
// ------------------------------------------------------------------------------------------------

//! Verfeinerung: Alles Dreiecke werden mittels des Bisektionsalgorithmus verfeinert
void Grid::refineAll()
{
    for ( TriangleIterator tIter = *this; tIter.notAtEnd(); ++tIter )
    {
        tIter->markForRefinement();
    }

    refineMarkedTriangles();
}
// ------------------------------------------------------------------------------------------------

//! Verfeinerung: Jedes markierte Dreieck wird mittels des Bisektionsalgorithmus verfeinert
void Grid::refineMarkedTriangles()
{
    for ( TriangleIterator tIter = *this; tIter.notAtEnd(); ++tIter )
    {
        if ( tIter->isMarkedForRefinement() ) refine( tIter->getIndex() );
    }
}
// ------------------------------------------------------------------------------------------------

//! Eigentliche Verfeinerung
//! \param  triangleToRefine    Globaler Index des zu verfeinernden Dreiecks
GlobalIndex Grid::refine( GlobalIndex triangleToRefine )
{
    LocalIndex l = getTriangle(triangleToRefine).getLongestEdgeIndex();
    DartIterator d( getTriangle(triangleToRefine), l );
    return refine(d);
}
// ------------------------------------------------------------------------------------------------

//! \brief results in at least one bisection of d.triangle along d.edge.
//!
//! modifies d.triangle such that a new node in the middle of d.edge
//! is inserted and d.triangle has d.node, the new point and the point
//! opposite to d.edge after the function's end.
//! The function returns the index of a new triangle that lies on the other half
//! of d.edge and is correctly connected with all triangles that were neighboured to
//! d.triangle before calling bisect().
//! If d.edge was not the longest edge of d.triangle, more bisections along other edges are made
//! before d.edge is cut in two. Indices of these inbetween-created triangles are not
//! returned.
GlobalIndex Grid::refine( const DartIterator& d)
{
    // first cut the neighbour on edge l in two triangles:
    GlobalIndex P_neu = addEdgeMidpoint(d);
    DartIterator d_prime = d;
    bool hasNeighbourOn_d_prime = d_prime.canFlipTriangle();
    GlobalIndex T_l_neu = IndexNotSet;

    if (hasNeighbourOn_d_prime)
    {
        d_prime.flipTriangle();
        T_l_neu = refineOnlyThis(d_prime, P_neu);
    }
    else{
      handlePeriodicCase( d );
    }
    // create new triangle T_neu and connect with neighbours
    GlobalIndex T_neu = refineOnlyThis(d, P_neu);

    if (hasNeighbourOn_d_prime)
    {
        // set d_prime to edge which lies in T_l_neu, opposite to edge d2.edge of T_neu
        d_prime.flipNode();
        d_prime.flipEdge();
        // flipTriangle() here will always work, as we have just created this
        // triangle. Thus, no canFlipTriangle() call is needed.
        d_prime.flipTriangle();
        d_prime.flipEdge();
        // set neighbour GlobalIndex in T_l_neu
        d_prime.getTriangle().setNeighbourIndex(d_prime.getLocalEdgeIndex(), T_neu);

        // move d1 to T_neu
        DartIterator d1 = d; d1.flipNode(); d1.flipEdge();
        d1.flipTriangle();
        d1.flipEdge();
        // set neighbour in T_neu
        d1.getTriangle().setNeighbourIndex(d1.getLocalEdgeIndex(), T_l_neu);
    }
    return T_neu;
}
// ------------------------------------------------------------------------------------------------

GlobalIndex Grid::refineOnlyThis( const DartIterator& d, GlobalIndex midpoint)
{
    // NOTE: in comments for this function, we write T for *this.
    LocalIndex l = d.getTriangle().getLongestEdgeIndex();
    if (l != d.getLocalEdgeIndex())
    {
        DartIterator d_l(d.getTriangle(), l);
        LocalIndex commonNode = d.getCommonNodeLocalIndex(d_l);
        DartIterator d_prime(d.getTriangle(), l, commonNode);
        // bisect T along the longest edge
        refine(d_prime);
        // because d_prime.node also lies on d.edge, T is now modifies in such a way
        // that d.edge has remained unchanged. Proceed by recursion until d.edge is
        // the longest edge in T.
        return refineOnlyThis(d, midpoint);
    }

    // create DartIterators to all edges, starting from l
    DartIterator d1 = d;  d1.flipNode(); d1.flipEdge();
    DartIterator d2 = d1; d2.flipNode(); d2.flipEdge();

    // create new triangle T_neu and connect with neighbours
    GlobalIndex T_neuIndex = addTriangle(midpoint, d1.getGlobalNodeIndex(), d2.getGlobalNodeIndex());
    Triangle & T_neu = getTriangle(T_neuIndex);
    T_neu.setNeighbourIndex(0, d1.getNextTriangleIndex());
    T_neu.setNeighbourIndex(1, d.getGlobalTriangleIndex());
    T_neu.setNeighbourIndex(2, d.getNextTriangleIndex());

    // set neighbour GlobalIndex in triangle lying opposite to d1.edge to T_neu
    // or update triangle map in the periodic case
    DartIterator d1_prime = d1;
    if (d1_prime.canFlipTriangle())
    {
        d1_prime.flipTriangle();
        d1_prime.getTriangle().setNeighbourIndex(d1_prime.getLocalEdgeIndex(), T_neuIndex);
    }
    else{
      updateTriangleMap( d1.getGlobalTriangleIndex(), d1.getLocalEdgeIndex(), T_neuIndex );
    }
    
    // set node at d1.node to newly created node:
    d1.getTriangle().setNodeIndex(d1.getLocalNodeIndex(), midpoint);
    d1.getTriangle().computeTransformation();
    
    // connect ourselves (i. e., connect T) to T_neu:
    d1.getTriangle().setNeighbourIndex(d1.getLocalEdgeIndex(), T_neuIndex);
    
    d1.getTriangle().unsetRefinementMarker();
    return T_neuIndex;
}
// ------------------------------------------------------------------------------------------------

GlobalIndex Grid::addEdgeMidpoint( const DartIterator& d)
{
    return addNode(d.getTriangle().getEdgeMidpoint(d.getLocalEdgeIndex()));
}
// ------------------------------------------------------------------------------------------------


//! Ausgabefunktion mit <<-Operator
ostream & operator<<(ostream & str, const Grid & grid)
{
    str << "Printing grid " << &grid << ":" << endl << endl
        << "Nodes:" << endl;

    for (Grid::NodeConstIterator iter = grid; iter.notAtEnd(); ++iter)
        str << *iter;

    str << endl
        << "Triangles:" << endl;

    for (Grid::TriangleConstIterator iter = grid; iter.notAtEnd(); ++iter)
        str << *iter;

    str << endl << endl;
    return str;
}
// ------------------------------------------------------------------------------------------------

//! \brief Erstellt Plot des Gitters
//! \param  filename    Dateiname fuer die Ausgabe (Standardeinstellung: output)
//! \param  format      Das Ausgabeformat (Standardeinstellung: POSTSCRIPT)
//!                     - GNUPLOT: Zur Betrachtung mittels 'gnuplot filename.txt -'
//!                     - JPG: Erstellt jpg-Bild
//!                     - MATLAB
//!                     - POSTSCRIPT: Erstellt eps-Datei
//!                     - PLY: Datei zur Betrachtung mit vtkfox
//! \param  neighbours  Stellt zusaetzlich Nachbarschaftsbeziehungen grafisch dar (Standardeinstellung: false)
//! \param  marked      Faerbt Dreiecke ein, die zur Verfeinerung markiert wurden (Standardeinstellung: false)
//!
void Grid::plot( string filename, FormatType format, bool neighbours, bool marked ) const
{
    fstream file;

    switch ( format )
    {
        case MATLAB:
        {
            // print triangulation into three files used by MatLab to display the triangulation:
            // filename + "_x.dat" and filename + "_y.dat" for x- and y-coordinates of the
            // grid nodes, and filename + "_tri.dat" for a 3-by-m-matrix containing the global
            // node indices. The triangulation can be displayed in MatLab loading all three files
            // and using "triplot(tri,x,y)".
            filename = filename.substr( 0, filename.rfind(".") );

            ofstream x_file((filename + "_x.dat").c_str()),
                     y_file((filename + "_y.dat").c_str());

            for (NodeConstIterator nIter = *this; nIter.notAtEnd(); ++nIter)
            {
                x_file << (*nIter)[0] << endl;
                y_file << (*nIter)[1] << endl;
            }
            x_file.close();
            y_file.close();

            ofstream tri_file((filename + "_tri.dat").c_str());
            for (TriangleConstIterator tIter = *this; tIter.notAtEnd(); ++tIter)
                tri_file << tIter->getNodeIndex(0) + 1 << "\t"
                         << tIter->getNodeIndex(1) + 1 << "\t"
                         << tIter->getNodeIndex(2) + 1 << endl;
            tri_file.close();
            break;
        }

        case GNUPLOT:
        {
            filename = filename.substr( 0, filename.rfind(".") );
            file.open( (filename+".txt").c_str(), fstream::out );
            file << "set title 'Dreiecksnetz'" << endl;
            file << "set size square" << endl;
            file << "unset border" << endl;
            file << "set xtics nomirror" << endl;
            file << "set ytics nomirror" << endl;
            file << "plot ";
            if ( marked )     file << "'" << filename << "_m.txt' notitle with filledcurves lw 1 lc rgb \"red\",";
            if ( neighbours ) file << "'" << filename << "_n.txt' notitle with lines lw 1 lc rgb \"green\",";
            file << "'" << filename << "_t.txt' notitle with lines lw 2 lc rgb \"black\"" << endl;
            file.close();

            file.open( (filename+"_t.txt").c_str(), fstream::out );
            if ( file.is_open() )
            {
                for ( TriangleConstIterator iter ( (*this) ); iter.notAtEnd(); ++iter )
                {
                    file << iter->getNode(0)[0] << " " << iter->getNode(0)[1] << endl;
                    file << iter->getNode(1)[0] << " " << iter->getNode(1)[1] << endl;
                    file << iter->getNode(2)[0] << " " << iter->getNode(2)[1] << endl;
                    file << iter->getNode(0)[0] << " " << iter->getNode(0)[1] << endl << endl;
                }
                file.close();
                cout << "To see the plot use 'gnuplot " << filename << ".txt -'" << endl;
            }
            else
            {
                cerr << "Error opening file " << filename << endl;
            }

            if ( neighbours )
            {
                file.open( (filename+"_n.txt").c_str(), fstream::out );
                if ( file.is_open() )
                {
                    for ( TriangleConstIterator iter ( (*this) ); iter.notAtEnd(); ++iter )
                    {
                        SmallVector<2> midpoint  = iter->getCenter();
                        for (LocalIndex i = 0; i < 3; ++i)
                            if (iter->getNeighbourIndex(i) != IndexNotSet)
                            {
                                SmallVector<2> otherMidpoint = iter->getNeighbour(i).getCenter();
                                file << midpoint[0] << " " << midpoint[1] << endl;
                                file << otherMidpoint[0] << " " << otherMidpoint[1] << endl << endl;
                            }
                    }
                    file.close();
                }
                else
                {
                    cerr << "Error opening file " << filename << endl;
                }
            }

            if ( marked )
            {
                file.open( (filename+"_m.txt").c_str(), fstream::out );
                if ( file.is_open() )
                {
                    for ( TriangleConstIterator iter ( (*this) ); iter.notAtEnd(); ++iter )
                    {
                        if ( (*iter).isMarkedForRefinement() )
                        {
                            file << iter->getNode(0)[0] << " " << iter->getNode(0)[1] << endl;
                            file << iter->getNode(1)[0] << " " << iter->getNode(1)[1] << endl;
                            file << iter->getNode(2)[0] << " " << iter->getNode(2)[1] << endl;
                            file << iter->getNode(0)[0] << " " << iter->getNode(0)[1] << endl << endl;
                        }
                    }
                    file.close();
                }
                else
                {
                    cerr << "Error opening file " << filename << endl;
                }
            }

            break;
        }

        case POSTSCRIPT:
        {
            filename = filename.substr( 0, filename.rfind(".") ) + ".eps";
            stringstream ss; ss << getpid();
            string tmpfile = ".gp_" + ss.str() + ".tmp";
            file.open( tmpfile.c_str(), fstream::out );
            if ( file.is_open() )
            {
                file << "set terminal postscript enhanced color eps" << endl;
                if ( filename == "output" ) file << "set output \"output.eps\"" << endl;
                else file << "set output \"" << filename << "\"" << endl;
                file << "set title '{/*2 Dreiecksnetz}'" << endl;
                file << "set size square 1, 1.4285715" << endl;
                file << "unset border" << endl;
                file << "set xtics nomirror" << endl;
                file << "set ytics nomirror" << endl;
                file << "plot ";
                if ( marked )     file << "'-' notitle with filledcurves lw 1 lc rgb \"red\",";
                if ( neighbours ) file << "'-' notitle with lines lw 1 lc rgb \"green\",";
                file << "'-' notitle with lines lt 1 lw 2 lc rgb \"black\"" << endl;

                if ( marked ) {
                    for ( TriangleConstIterator iter ( (*this) ); iter.notAtEnd(); ++iter )
                    {
                        if ( (*iter).isMarkedForRefinement() )
                        {
                            file << iter->getNode(0)[0] << " " << iter->getNode(0)[1] << endl;
                            file << iter->getNode(1)[0] << " " << iter->getNode(1)[1] << endl;
                            file << iter->getNode(2)[0] << " " << iter->getNode(2)[1] << endl;
                            file << iter->getNode(0)[0] << " " << iter->getNode(0)[1] << endl << endl;
                        }
                    }
                    file << "e" << endl;
                }

                if ( neighbours ) {
                    for ( TriangleConstIterator iter ( (*this) ); iter.notAtEnd(); ++iter )
                    {
                        SmallVector<2> midpoint  = iter->getCenter();
                        for (LocalIndex i = 0; i < 3; ++i)
                            if (iter->getNeighbourIndex(i) != IndexNotSet)
                            {
                                SmallVector<2> otherMidpoint = iter->getNeighbour(i).getCenter();
                                file << midpoint[0] << " " << midpoint[1] << endl;
                                file << otherMidpoint[0] << " " << otherMidpoint[1] << endl << endl;
                            }
                    }
                    file << "e" << endl;
                }

                for ( TriangleConstIterator iter ( (*this) ); iter.notAtEnd(); ++iter )
                {
                    file << iter->getNode(0)[0] << " " << iter->getNode(0)[1] << endl;
                    file << iter->getNode(1)[0] << " " << iter->getNode(1)[1] << endl;
                    file << iter->getNode(2)[0] << " " << iter->getNode(2)[1] << endl;
                    file << iter->getNode(0)[0] << " " << iter->getNode(0)[1] << endl << endl;
                }
                file << "e" << endl;

                file.close();
                system( (GNUPLOT_PATH + tmpfile).c_str() );
                remove( tmpfile.c_str() );
            }
            else
            {
                cerr << "Error opening file " << filename << endl;
            }
            break;
        }

        case JPG:
        {
            filename = filename.substr( 0, filename.rfind(".") ) + ".jpg";
            stringstream ss; ss << getpid();
            string tmpfile = ".gp_" + ss.str() + ".tmp";
            file.open( tmpfile.c_str(), fstream::out );
            if ( file.is_open() )
            {
                file << "set terminal jpeg large size 800,800 enhanced" << endl;
                if ( filename == "output" ) file << "set output \"output.jpg\"" << endl;
                else file << "set output \"" << filename << "\"" << endl;
                file << "set title 'Dreiecksnetz'" << endl;
                file << "unset border" << endl;
                file << "set xtics nomirror" << endl;
                file << "set ytics nomirror" << endl;
                file << "plot ";
                if ( marked )     file << "'-' notitle with filledcurves lw 1 lc rgb \"red\",";
                if ( neighbours ) file << "'-' notitle with lines lw 1 lc rgb \"green\",";
                file << "'-' notitle with lines lw 2 lc rgb \"black\"" << endl;

                if ( marked ) {
                    for ( TriangleConstIterator iter ( (*this) ); iter.notAtEnd(); ++iter )
                    {
                        if ( (*iter).isMarkedForRefinement() )
                        {
                            file << iter->getNode(0)[0] << " " << iter->getNode(0)[1] << endl;
                            file << iter->getNode(1)[0] << " " << iter->getNode(1)[1] << endl;
                            file << iter->getNode(2)[0] << " " << iter->getNode(2)[1] << endl;
                            file << iter->getNode(0)[0] << " " << iter->getNode(0)[1] << endl << endl;
                        }
                    }
                    file << "e" << endl;
                }

                if ( neighbours ) {
                    for ( TriangleConstIterator iter ( (*this) ); iter.notAtEnd(); ++iter )
                    {
                        SmallVector<2> midpoint  = iter->getCenter();
                        for (LocalIndex i = 0; i < 3; ++i)
                            if (iter->getNeighbourIndex(i) != IndexNotSet)
                            {
                                SmallVector<2> otherMidpoint = iter->getNeighbour(i).getCenter();
                                file << midpoint[0] << " " << midpoint[1] << endl;
                                file << otherMidpoint[0] << " " << otherMidpoint[1] << endl << endl;
                            }
                    }
                    file << "e" << endl;
                }

                for ( TriangleConstIterator iter ( (*this) ); iter.notAtEnd(); ++iter )
                {
                    file << iter->getNode(0)[0] << " " << iter->getNode(0)[1] << endl;
                    file << iter->getNode(1)[0] << " " << iter->getNode(1)[1] << endl;
                    file << iter->getNode(2)[0] << " " << iter->getNode(2)[1] << endl;
                    file << iter->getNode(0)[0] << " " << iter->getNode(0)[1] << endl << endl;
                }
                file << "e" << endl;

                file.close();
                system( (GNUPLOT_PATH + tmpfile).c_str() );
                remove( tmpfile.c_str() );
            }
            else
            {
                cerr << "Error opening file " << filename << endl;
            }
            break;
        }

        case PLY:
        {
            filename = filename.substr( 0, filename.rfind(".") ) + ".ply";
            file.open( filename.c_str(), fstream::out );
            if ( file.is_open() )
            {
                file << "ply\n"
                     << "format ascii 1.0\n"
                     << "element vertex " << _nodeList.size() << endl
                     << "property float x\n"
                     << "property float y\n"
                     << "property float z\n"
                     << "element face "   << _triangleList.size() << endl
                     << "property list uchar int vertex_index\n"
                     << "end_header\n";
                for ( NodeConstIterator iter ( (*this) ); iter.notAtEnd(); ++iter )
                {
                    file << (*iter)[0] << " " << (*iter)[1] << " " << 0.0 << endl;
                }
                for ( TriangleConstIterator iter ( (*this) ); iter.notAtEnd(); ++iter )
                {
                    file << iter->getNodeIndex(0) << " " << iter->getNodeIndex(1) << " " << iter->getNodeIndex(2) << endl;
                }
            }
            else
            {
                cerr << "Error opening file " << filename << endl;
            }
            break;
        }
    }

    return;
}
// ------------------------------------------------------------------------------------------------

//! \brief Erstellt Plot der Loesung
//! \param  sol         Loesungsvektor (muss soviele Eintraege wie Knoten im Gitter haben!)
//! \param  filename    Dateiname fuer die Ausgabe (Standardeinstellung: output)
//! \param  format      Das Ausgabeformat (Standardeinstellung: POSTSCRIPT)
//!                     - GNUPLOT: Zur Betrachtung mittels 'gnuplot gnuplot.txt -'
//!                     - JPG: Erstellt jpg-Bild
//!                     - MATLAB
//!                     - POSTSCRIPT: Erstellt eps-Datei
//!                     - PLY: Datei zur Betrachtung mit vtkfox
//! \param  flat        Schaltet in gnuplot auf 2D Projektion von oben (Standardeinstellung: true)
//! \param  z_min       Minimaler z-Wert im Plot (Standardeinstellung: 0)
//! \param  z_max       Maximaler z-Wert im Plot (Standardeinstellung: 0)
//!                     In der Standardeinstellung wird der Wertebereich automatisch durch gnuplot festgelegt!
//! \bug                Im 3D-Modus funktioniert leider keine Entfernung verdeckter Bereiche!
//!
void Grid::plotSolution( Vector & sol, string filename, FormatType format, bool flat, double z_min, double z_max ) const
{
    if ( sol.size() < static_cast<size_t>( getNumberOfNodes() ) )  {
        cerr << "Solution Vector does not have enough entries!" << endl;
        return;
    }

    fstream file;

    switch ( format )
    {
        case MATLAB:
        {
            filename = filename.substr( 0, filename.rfind(".") );
            plot( filename, MATLAB );

            ofstream data_file((filename + ".dat").c_str());
            for ( size_t i = 0; i < sol.size(); ++i )
                data_file << sol[i] << endl;
            data_file.close();

            string m_filename = string("plot_") + filename + ".m";
            ofstream m_file(m_filename.c_str());
            string tri_file = filename + "_tri",
                   x_file   = filename + "_x",
                   y_file   = filename + "_y";
            m_file << "load '" << tri_file << ".dat';" << endl
                   << "load '" << x_file << ".dat';" << endl
                   << "load '" << y_file << ".dat';" << endl
                   << "load '" << filename << ".dat';" << endl
                   << "trisurf(" << tri_file << "," << x_file << "," << y_file << "," << filename << ");";
            m_file.close();

            break;
        }

        case GNUPLOT:
        {
            filename = filename.substr( 0, filename.rfind(".") ) + ".txt";
            file.open( "gnuplot.txt", fstream::out );
            file << "set size square" << endl;
            file << "set pm3d interpolate 10,10 implicit" << ( flat ? " map" : "") << endl;
            if ( !flat )
            {
                file << "set view 60,45" << endl;
                file << "set grid back" << endl;
            }
            file << "set palette model HSV" << endl;
            file << "set palette defined (0 0.67 1 0.7, 0.33 0.5 1 0.8, 0.67 0.166 1.0 0.9, 1 0 1 1)" << endl;
            file << "set colorbox user size .02,.6" << endl;
            file << "unset border" << endl;
            file << "set xtics nomirror" << endl;
            file << "set ytics nomirror" << endl;
            if ( fabs( z_max - z_min ) > 1e-15 )
            {
                if ( !flat ) file << "set zrange [" << z_min << ":" << z_max << "]" << endl;
                file << "set cbrange [" << z_min << ":" << z_max << "]" << endl;
            }
            file << "splot '" << filename << "' notitle with dots" << endl;
            file.close();

            file.open( filename.c_str(), fstream::out );
            if ( file.is_open() )
            {
                for ( TriangleConstIterator iter ( (*this) ); iter.notAtEnd(); ++iter )
                {
                    file << iter->getNode(0)[0] << " " << iter->getNode(0)[1] << " " << sol[iter->getNodeIndex(0)] << endl;
                    file << iter->getNode(1)[0] << " " << iter->getNode(1)[1] << " " << sol[iter->getNodeIndex(1)] << endl << endl;
                    file << iter->getNode(0)[0] << " " << iter->getNode(0)[1] << " " << sol[iter->getNodeIndex(0)] << endl;
                    file << iter->getNode(2)[0] << " " << iter->getNode(2)[1] << " " << sol[iter->getNodeIndex(2)] << endl << endl << endl;
                }
                file.close();
                cout << "To see the plot use 'gnuplot gnuplot.txt -'" << endl << endl;
            }
            else
            {
                cerr << "Error opening file " << filename << endl;
            }
            break;
        }

        case POSTSCRIPT:
        {
            filename = filename.substr( 0, filename.rfind(".") ) + ".eps";
            stringstream ss; ss << getpid();
            string tmpfile = ".gp_" + ss.str() + ".tmp";
            file.open( tmpfile.c_str(), fstream::out );
            if ( file.is_open() )
            {
                file << "set terminal postscript enhanced color eps" << endl;
                file << "set output \"" << filename << "\"" << endl;
                file << "set size square 1, 1.4285715" << endl;
                file << "set pm3d interpolate 10,10 implicit" << ( flat ? " map" : "") << endl;
                if ( !flat )
                {
                    file << "set view 60,45" << endl;
                    file << "set grid back" << endl;
                }
                file << "set palette model HSV" << endl;
                file << "set palette defined (0 0.67 1 0.7, 0.33 0.5 1 0.8, 0.67 0.166 1.0 0.9, 1 0 1 1)" << endl;
                file << "set colorbox user size .02,.6" << endl;
                file << "unset border" << endl;
                file << "set xtics nomirror" << endl;
                file << "set ytics nomirror" << endl;
                if ( fabs( z_max - z_min ) > 1e-15 )
                {
                    if ( !flat ) file << "set zrange [" << z_min << ":" << z_max << "]" << endl;
                    file << "set cbrange [" << z_min << ":" << z_max << "]" << endl;
                }
                file << "splot '-' notitle with dots" << endl;

                for ( TriangleConstIterator iter ( (*this) ); iter.notAtEnd(); ++iter )
                {
                    file << iter->getNode(0)[0] << " " << iter->getNode(0)[1] << " " << sol[iter->getNodeIndex(0)] << endl;
                    file << iter->getNode(1)[0] << " " << iter->getNode(1)[1] << " " << sol[iter->getNodeIndex(1)] << endl << endl;
                    file << iter->getNode(0)[0] << " " << iter->getNode(0)[1] << " " << sol[iter->getNodeIndex(0)] << endl;
                    file << iter->getNode(2)[0] << " " << iter->getNode(2)[1] << " " << sol[iter->getNodeIndex(2)] << endl << endl << endl;
                }
                file << "e" << endl;

                file.close();
                system( (GNUPLOT_PATH + tmpfile).c_str() );
                remove( tmpfile.c_str() );
            }
            else
            {
                cerr << "Error opening file " << filename << endl;
            }
            break;
        }

        case JPG:
        {
            filename = filename.substr( 0, filename.rfind(".") ) + ".jpg";
            stringstream ss; ss << getpid();
            string tmpfile = ".gp_" + ss.str() + ".tmp";
            file.open( tmpfile.c_str(), fstream::out );
            if ( file.is_open() )
            {
                file << "set terminal jpeg large size 800,800 enhanced" << endl;
                file << "set output \"" << filename << "\"" << endl;
                file << "set pm3d interpolate 10,10 implicit" << ( flat ? " map" : "") << endl;
                if ( !flat )
                {
                    file << "set view 60,45" << endl;
                    file << "set grid back" << endl;
                }
                file << "set palette model HSV" << endl;
                file << "set palette defined (0 0.67 1 0.7, 0.33 0.5 1 0.8, 0.67 0.166 1.0 0.9, 1 0 1 1)" << endl;
                file << "set colorbox user size .02,.6" << endl;
                file << "unset border" << endl;
                file << "set xtics nomirror" << endl;
                file << "set ytics nomirror" << endl;
                if ( fabs( z_max - z_min ) > 1e-15 )
                {
                    if ( !flat ) file << "set zrange [" << z_min << ":" << z_max << "]" << endl;
                    file << "set cbrange [" << z_min << ":" << z_max << "]" << endl;
                }
                file << "splot '-' notitle with dots" << endl;

                for ( TriangleConstIterator iter ( (*this) ); iter.notAtEnd(); ++iter )
                {
                    file << iter->getNode(0)[0] << " " << iter->getNode(0)[1] << " " << sol[iter->getNodeIndex(0)] << endl;
                    file << iter->getNode(1)[0] << " " << iter->getNode(1)[1] << " " << sol[iter->getNodeIndex(1)] << endl << endl;
                    file << iter->getNode(0)[0] << " " << iter->getNode(0)[1] << " " << sol[iter->getNodeIndex(0)] << endl;
                    file << iter->getNode(2)[0] << " " << iter->getNode(2)[1] << " " << sol[iter->getNodeIndex(2)] << endl << endl << endl;
                }
                file << "e" << endl;

                file.close();
                system( (GNUPLOT_PATH + tmpfile).c_str() );
                remove( tmpfile.c_str() );
            }
            else
            {
                cerr << "Error opening file " << filename << endl;
            }
            break;
        }

        case PLY:
        {
            filename = filename.substr( 0, filename.rfind(".") ) + ".ply";
            file.open( filename.c_str(), fstream::out );
            if ( file.is_open() )
            {
                file << "ply\n"
                     << "format ascii 1.0\n"
                     << "element vertex " << _nodeList.size() << endl
                     << "property float x\n"
                     << "property float y\n"
                     << "property float z\n"
                     << "property float generic_scalar_values\n"
                     << "element face " << _triangleList.size() << endl
                     << "property list uchar int vertex_index\n"
                     << "end_header\n";
                for ( NodeConstIterator iter ( (*this) ); iter.notAtEnd(); ++iter )
                {
                    file << (*iter)[0] << " " << (*iter)[1] << " " << sol[ iter.getIndex() ] << " " << sol[ iter.getIndex() ] << endl;
                }
                for ( TriangleConstIterator iter ( (*this) ); iter.notAtEnd(); ++iter )
                {
                    file << iter->getNodeIndex(0) << " " << iter->getNodeIndex(1) << " " << iter->getNodeIndex(2) << endl;
                }
            }
            else
            {
                cerr << "Error opening file " << filename << endl;
            }
            break;
        }

        default:
        {
            cerr << "Case " << format << " not handled!" << endl;
        }
    }

    return;
}
// ------------------------------------------------------------------------------------------------

//! \brief Erstellt Plot der Loesung
//! \param  sol         Loesungsvektor (muss soviele Eintraege wie Knoten im Gitter haben!)
//! \param  def         Vektor fuer Deformation (zwei mal soviele Eintraege wie Knoten im Gitter)
//! \param  filename    Dateiname fuer die Ausgabe (Standardeinstellung: output)
//! \param  format      Das Ausgabeformat (Standardeinstellung: JPG)
//!                     - JPG: Erstellt jpg-Bild
//! \param  z_min       Minimaler z-Wert im Plot (Standardeinstellung: 0)
//! \param  z_max       Maximaler z-Wert im Plot (Standardeinstellung: 0)
//!                     In der Standardeinstellung wird der Wertebereich automatisch durch gnuplot festgelegt!
//!
void Grid::plotSolutionDeformed( const Vector & sol, const Vector & def, string filename, FormatType format, double z_min, double z_max ) const
{
    if ( sol.size() < static_cast<size_t>( getNumberOfNodes() ) )  {
      cerr << "Solution Vector does not have enough entries!" << endl;
      return;
    }
    if ( def.size() < 2*static_cast<size_t>( getNumberOfNodes() ) )  {
      cerr << "Deformation Vector does not have enough entries!" << endl;
      return;
    }

    const int N = getNumberOfNodes();
    fstream file;

    switch ( format )
    {
        case JPG:
        {
            filename = filename.substr( 0, filename.rfind(".") ) + ".jpg";
            stringstream ss; ss << getpid();
            string tmpfile = ".gp_" + ss.str() + ".tmp";
            file.open( tmpfile.c_str(), fstream::out );
            if ( file.is_open() )
            {
                file << "set terminal jpeg large size 800,800 enhanced" << endl;
                file << "set output \"" << filename << "\"" << endl;
                file << "set pm3d interpolate 10,10 implicit map" << endl;
                file << "set palette model HSV" << endl;
                file << "set palette defined (0 0.67 1 0.7, 0.33 0.5 1 0.8, 0.67 0.166 1.0 0.9, 1 0 1 1)" << endl;
                file << "set colorbox user size .02,.6" << endl;
                file << "unset border" << endl;
                file << "set xtics nomirror" << endl;
                file << "set ytics nomirror" << endl;
                if ( fabs( z_max - z_min ) > 1e-15 ) file << "set cbrange [" << z_min << ":" << z_max << "]" << endl;
                file << "splot '-' notitle with dots" << endl;

                for ( TriangleConstIterator iter ( (*this) ); iter.notAtEnd(); ++iter )
                {
                  file << iter->getNode(0)[0]+def[iter->getNodeIndex(0)] << " " << iter->getNode(0)[1]+def[N+iter->getNodeIndex(0)] << " " << sol[iter->getNodeIndex(0)] << endl;
                  file << iter->getNode(1)[0]+def[iter->getNodeIndex(1)] << " " << iter->getNode(1)[1]+def[N+iter->getNodeIndex(1)] << " " << sol[iter->getNodeIndex(1)] << endl << endl;
                  file << iter->getNode(0)[0]+def[iter->getNodeIndex(0)] << " " << iter->getNode(0)[1]+def[N+iter->getNodeIndex(0)] << " " << sol[iter->getNodeIndex(0)] << endl;
                  file << iter->getNode(2)[0]+def[iter->getNodeIndex(2)] << " " << iter->getNode(2)[1]+def[N+iter->getNodeIndex(2)] << " " << sol[iter->getNodeIndex(2)] << endl << endl << endl;
                }
                file << "e" << endl;

                file.close();
                system( (GNUPLOT_PATH + tmpfile).c_str() );
                remove( tmpfile.c_str() );
            }
            else
            {
                cerr << "Error opening file " << filename << endl;
            }
            break;
        }

        default:
        {
            cerr << "Case " << format << " not handled!" << endl;
        }
  }
  return;
}

bool Grid::handlePeriodicCase( const DartIterator& ){ return false; }
void Grid::updateTriangleMap( const GlobalIndex, const LocalIndex, const GlobalIndex ){}

// ------------------------------------------------------------------------------------------------

PeriodicSquare::PeriodicSquare( int GridDepth )  {
  this->Grid::operator=( createRectangularGrid( 0, 0, 1, 1, GridDepth ) );  
  assembleTriangleMap();
}

void PeriodicSquare::assembleTriangleMap()
{
    _tmap.clear();

     // run over boundary and store all edges and triangles
    bool cornerSeen = false;
    int nextCornerIndex = IndexNotSet;    
    SmallVector2<> olddiff (1.,1.), newdiff (1.,1.);
    // container to store boundary edges and triangles
    vector<GlobalIndex> tindices, tauxVec;
    vector<LocalIndex> eindices, eauxVec;

    for (Grid::BoundaryConstIterator bIter = *this; bIter.notAtEnd(); ++bIter)
    {
      // looking for a corner (x,y) with x=y
      if ( abs( bIter.getStart()[0] - bIter.getStart()[1] ) < 1e-15 ) cornerSeen = true;
        if (cornerSeen) {
            tindices.push_back( bIter.getTriangleIndex() );
            eindices.push_back( bIter.getEdgeIndex() );
            if ( nextCornerIndex == IndexNotSet )  {
                newdiff = bIter.getStart();
                newdiff-= bIter.getEnd();
                // check for right angle
                if ( abs( newdiff * olddiff ) < 1e-15 ) nextCornerIndex = tindices.size()-1;
                olddiff = newdiff;
            }
        }
        // before the first corner was seen all indices are stored in auxiliary vectors
        else  {
          tauxVec.push_back( bIter.getTriangleIndex() );
          eauxVec.push_back( bIter.getEdgeIndex() );
        }
    }

    // perform shifting, i.e. add indices by auxiliary indices
    tindices.insert( tindices.end(), tauxVec.begin(), tauxVec.end() );
    eindices.insert( eindices.end(), eauxVec.begin(), eauxVec.end() );


    // now create pairs
    const unsigned int numBdryNodes = tindices.size();
    assert( numBdryNodes % 2 == 0);
    const unsigned int numNodesFirstDir  = nextCornerIndex+1;
    const unsigned int numNodesSecondDir = numBdryNodes/2-numNodesFirstDir;

    // different to the node map in periodicBoundaryMask, _tmap should be bidirectional
    for ( unsigned int i = 0; i < numNodesFirstDir-1; ++i )
    {
      _tmap[make_pair( tindices[i], eindices[i] )] = make_pair( tindices[tindices.size()-2-numNodesSecondDir-i], eindices[tindices.size()-2-numNodesSecondDir-i] );
      _tmap[make_pair( tindices[tindices.size()-2-numNodesSecondDir-i], eindices[tindices.size()-2-numNodesSecondDir-i] )] = make_pair( tindices[i], eindices[i] );
    }

    for ( unsigned int i = 0; i < numNodesSecondDir+1; ++i )
    {
      _tmap[make_pair( tindices[numNodesFirstDir-1+i], eindices[numNodesFirstDir-1+i] )] = make_pair( tindices[tindices.size()-1-i], eindices[tindices.size()-1-i] );
      _tmap[make_pair( tindices[tindices.size()-1-i], eindices[tindices.size()-1-i] )] = make_pair( tindices[numNodesFirstDir-1+i], eindices[numNodesFirstDir-1+i] );
    }

}

bool PeriodicSquare::handlePeriodicCase( const DartIterator& d )
{
    TriangleMapType::const_iterator tmapIter = _tmap.find( make_pair(d.getGlobalTriangleIndex(),d.getLocalEdgeIndex()) );
    // this assertion is really necessary as we are definitely facing problems if this is no true!
    assert( tmapIter != _tmap.end() );

    // set d_other to corresponding triangle, flipTriangle won't work
    int otherEdgeIndex = (tmapIter->second).second;
    int otherNodeIndex = (otherEdgeIndex+1)%3;
    DartIterator d_other( this->getTriangle( (tmapIter->second).first ), otherEdgeIndex, otherNodeIndex );

    GlobalIndex P_clone = addEdgeMidpoint(d_other);
    this->refineOnlyThis(d_other, P_clone);

    return true;
}

void PeriodicSquare::updateTriangleMap( const GlobalIndex triangleIndex, const LocalIndex edgeIndex, const GlobalIndex newTriangleIndex )
{
    TriangleMapType::iterator tmapIter = _tmap.find( make_pair( triangleIndex, edgeIndex ) );

    // this assertion should not be active until there is another updating of the triangle map when a boundary edge is bisected
    // assert( tmapIter != _tmap.end() );     
    // The refinement works as long as a refined boundary edge is not touched again in the same refinement procedure!
    // However, this should not happen but it would be nice to have this update nevertheless!
    //! \bug Currently there is no update of the triangle map when a boundary edge is bisected!
    if( tmapIter == _tmap.end() )
      return;
    // it might happen that we encounter a triangle on the boundary which was created by bisection and did not inherit the periodic mapping
    // this is simply not implemented but should not cause trouble. Fingers crossed!

    // correct reverse mapping (affected boundary edge in new triangle has local index 0 by construction!)
    _tmap[ tmapIter->second ] = make_pair( newTriangleIndex, 0 );
    // correct forward mapping
    _tmap[ make_pair( newTriangleIndex, 0 ) ] = tmapIter->second;
    _tmap.erase( tmapIter );
}

void PeriodicSquare::refineMarkedTriangles()
{
  Grid::refineMarkedTriangles();
  assembleTriangleMap();
}
