/** Klasse zur Verwaltung einer Triangulierung
 *
 *  \file   grid.h
 *  \author Stefan W. von Deylen, Benedict Geihe
 *  \date   18.09.2008
 *
 */

#ifndef __grid_h
#define __grid_h

#include <string>
#include <fstream>
#include <map>

using namespace std;

class Grid;

#include "node.h"

class Triangle;

// ------------------------------------------------------------------------------------------------

template <typename GridType>
class TypeMapper {};

template <>
class TypeMapper<Grid>
{
    public:
        typedef Triangle TriangleType;
        typedef Node     NodeType;
};

template <>
class TypeMapper<const Grid>
{
    public:
        typedef const Triangle TriangleType;
        typedef const Node     NodeType;
};

// ------------------------------------------------------------------------------------------------

typedef map< pair<GlobalIndex,LocalIndex>, pair<GlobalIndex,LocalIndex> > TriangleMapType;
typedef enum { GNUPLOT, POSTSCRIPT, JPG, MATLAB, PLY } FormatType;

#include "smallVector.h"
#include "triangle.h"
#include "../vecmat/vector.h"
#include "../fe/boundaryMasks.h"

//! Gitter-Klasse
class Grid
{
    public:
        // Iteratoren auf Knoten und Dreiecken
        typedef NodeIteratorTemplate<Node, Grid>                     NodeIterator;
        typedef NodeIteratorTemplate<const Node, const Grid>         NodeConstIterator;
        typedef TriangleIteratorTemplate<Triangle, Grid>             TriangleIterator;
        typedef TriangleIteratorTemplate<const Triangle, const Grid> TriangleConstIterator;
        typedef BoundaryIteratorTemplate<Node, Grid>                 BoundaryIterator;
        typedef BoundaryIteratorTemplate<const Node, const Grid>     BoundaryConstIterator;
        typedef DartIteratorTemplate<Grid>                           DartIterator;
        typedef DartIteratorTemplate<const Grid>                     DartConstIterator;

        // Konstruktoren, Zuweisung
        Grid();
        Grid(const Node & node1, const Node & node2, const Node & node3);
        Grid(const Grid & otherGrid);
        Grid & operator=(const Grid & otherGrid);

        // Konstruktion des Gitters
        GlobalIndex addNode( const Node & nodeToAdd );
        GlobalIndex addNodeAndTriangle( const Node & nodeToAdd, DartIterator existentEdge );
        GlobalIndex addTriangle( DartIterator existentEdge1, DartIterator existentEdge2 );
        GlobalIndex addTriangle( DartIterator existentEdge1, DartIterator existentEdge2, DartIterator existentEdge3 );
        GlobalIndex addTriangle( GlobalIndex node1, GlobalIndex node2, GlobalIndex node3 );
        static Grid createRectangularGrid(double pt1_x, double pt1_y, double pt2_x, double pt2_y, int cnt_refine);

        // Zugriff auf Knoten und Dreiecke
              Node & getNode(GlobalIndex globalNodeIndex);
        const Node & getNode(GlobalIndex globalNodeIndex) const;
              Triangle & getTriangle(GlobalIndex globalTriangleIndex);
        const Triangle & getTriangle(GlobalIndex globalTriangleIndex) const;
        unsigned int getNumberOfNodes() const;
        unsigned int getNumberOfElements() const;

        // Verfeinerung
        void refineAll();
        void refine4All();
        virtual void refineMarkedTriangles();

        // Plot-Funktion
        void plot( string filename = "output", FormatType format = POSTSCRIPT, bool neighbours = false, bool marked = false ) const;
        void plotSolution( Vector & sol,
                           string filename = "output",
                           FormatType format = POSTSCRIPT,
                           bool flat = true,
                           double z_min = 0., double z_max = 0. ) const;
        void plotSolutionDeformed( const Vector & sol,
                                   const Vector & def,
                                   string filename = "output",
                                   FormatType format = JPG,
                                   double z_min = 0., double z_max = 0. ) const;

    protected:
        // Knoten und Dreiecksliste
        NodeList     _nodeList;
        TriangleList _triangleList;

        // Hilfsfunktion
        void assertEdgeIsBoundary( const DartIterator& edge ) const;
        virtual bool handlePeriodicCase( const DartIterator& );
        virtual void updateTriangleMap( const GlobalIndex triangleIndex, const LocalIndex edgeIndex, const GlobalIndex newTriangleIndex );
	
        // internal refinement functions
        GlobalIndex refine(GlobalIndex triangleToRefine);
        GlobalIndex refine( const DartIterator& d);
        GlobalIndex refineOnlyThis( const DartIterator& d, GlobalIndex midpoint);
        GlobalIndex addEdgeMidpoint( const DartIterator& d);

        // Gestatte den Iteratoren Zugriff auf die geschuetzten Daten des Gitters
        friend class TriangleIteratorTemplate<Triangle, Grid>;
        friend class TriangleIteratorTemplate<const Triangle, const Grid>;
        friend class NodeIteratorTemplate<Node, Grid>;
        friend class NodeIteratorTemplate<const Node, const Grid>;
        friend class BoundaryIteratorTemplate<Node, Grid>;
        friend class BoundaryIteratorTemplate<const Node, const Grid>;
};

//!
class PeriodicSquare : public Grid {
  
  public:
    PeriodicSquare( int GridDepth = 0 );
    void refineMarkedTriangles();    
    
  protected:
    TriangleMapType  _tmap; 
    void assembleTriangleMap();
    bool handlePeriodicCase( const DartIterator& dart );
    void updateTriangleMap( const GlobalIndex triangleIndex, const LocalIndex edgeIndex, const GlobalIndex newTriangleIndex );
};


//! Ausgabefunktion mit <<-Operator
ostream & operator<<(ostream & str, const Grid & grid);

#endif
