/** Klasse fuer kleine Vektoren (Koordinaten etc.)
 *
 *  \file   smallVector.cpp
 *  \author Stefan W. von Deylen, Benedict Geihe
 *  \date   18.09.2008
 *
 */


#include "smallVector.h"

using namespace std;

//! Standard-Konstruktor
template < unsigned int Size, typename DataType >
SmallVector< Size, DataType >::SmallVector()
{
    setZero();
}
// ------------------------------------------------------------------------------------------------

//! Setze alle Werte auf 0
template < unsigned int Size, typename DataType >
void SmallVector< Size, DataType >::setZero()
{
    for ( unsigned int i = 0; i < Size; ++i )
    {
        (*this)[i] = 0;
    }
}
// ------------------------------------------------------------------------------------------------

//! Normalisiert den Vektor
template < unsigned int Size, typename DataType >
void SmallVector< Size, DataType >::normalize()
{
    DataType norm = static_cast<DataType>( this->norm_l2() );
    (*this) /= norm;
}
// ------------------------------------------------------------------------------------------------

//! l2-Norm ausgeben
template < unsigned int Size, typename DataType >
double SmallVector< Size, DataType >::norm_l2() const
{
    return sqrt( static_cast<double>( norm_l2q() ) );
}
// ------------------------------------------------------------------------------------------------

//! Quadrat der l2-Norm ausgeben
template < unsigned int Size, typename DataType >
DataType SmallVector< Size, DataType >::norm_l2q() const
{
    DataType normsqr = 0;
    for ( unsigned int i = 0; i < Size; ++i )
    {
        normsqr += (*this)[i] * (*this)[i];
    }
    return normsqr;
}
// ------------------------------------------------------------------------------------------------

//! l-unendlich-Norm ausgeben
template < unsigned int Size, typename DataType >
DataType SmallVector< Size, DataType >::norm_linfty() const
{
    DataType maxentry = 0;
    for ( unsigned int i = 1; i < Size; ++i )
    {
        if ( Abs( (*this)[i] ) > maxentry ) maxentry = Abs( (*this)[i] );
    }
    return maxentry;
}
// ------------------------------------------------------------------------------------------------

//! Zugriff lesend
template < unsigned int Size, typename DataType >
DataType & SmallVector< Size, DataType >::operator[] (unsigned int i)
{
    return _entry[i];
}
// ------------------------------------------------------------------------------------------------

//! Zugriff schreibend
template < unsigned int Size, typename DataType >
const DataType & SmallVector< Size, DataType >::operator[] (unsigned int i) const
{
    return _entry[i];
}
// ------------------------------------------------------------------------------------------------

//! *this = d * v setzen (d reell, v Vektor)
template < unsigned int Size, typename DataType >
void SmallVector< Size, DataType >::setToMultiple( const DataType d, const SmallVector< Size, DataType >& v )
{
    for ( unsigned int i = 0; i < Size; ++i )
    {
        (*this)[i] = v[i] * d;
    }
}
// ------------------------------------------------------------------------------------------------

//! *this += d * v berechnen (d reell, v Vektor)
template < unsigned int Size, typename DataType >
void SmallVector< Size, DataType >::addMultiple( const DataType d, const SmallVector< Size, DataType >& v )
{
    for ( unsigned int i = 0; i < Size; ++i )
    {
        (*this)[i] += v[i] * d;
    }
}
// ------------------------------------------------------------------------------------------------

//! *= Operator: Multipliziert jeden Eintrag mit Faktor r
template < unsigned int Size, typename DataType >
SmallVector< Size, DataType > & SmallVector< Size, DataType >::operator*=( const DataType r )
{
    for ( unsigned int i = 0; i < Size; ++i )
    {
        (*this)[i] *= r;
    }
    return *this;
}
// ------------------------------------------------------------------------------------------------

//! /= Operator: Dividiert jeden Eintrag durch Faktor r
template < unsigned int Size, typename DataType >
SmallVector< Size, DataType > & SmallVector< Size, DataType >::operator/=( const DataType r )
{
    for ( unsigned int i = 0; i < Size; ++i )
    {
        (*this)[i] /= r;
    }
    return *this;
}
// ------------------------------------------------------------------------------------------------

//! += Operator: Addiert einen anderen Vektor
template < unsigned int Size, typename DataType >
SmallVector< Size, DataType > & SmallVector< Size, DataType >::operator+=( const SmallVector< Size, DataType > & v )
{
    for ( unsigned int i = 0; i < Size; ++i )
    {
        (*this)[i] += v[i];
    }
    return *this;
}
// ------------------------------------------------------------------------------------------------

//! -= Operator: Subtrahiert einen anderen Vektor
template < unsigned int Size, typename DataType >
SmallVector< Size, DataType > & SmallVector< Size, DataType >::operator-=(const SmallVector< Size, DataType > & v)
{
    for ( unsigned int i = 0; i < Size; ++i )
    {
        (*this)[i] -= v[i];
    }
    return *this;
}
// ------------------------------------------------------------------------------------------------

//! * Operator: Berechnet das Skalarprodukt zweier Vektoren
template < unsigned int Size, typename DataType >
DataType SmallVector< Size, DataType >::operator*( const SmallVector< Size, DataType > & v ) const
{
    DataType ret = 0;
    for ( unsigned int i = 0; i < Size; ++i )
    {
        ret += (*this)[i] * v[i];
    }
    return ret;
}
// ------------------------------------------------------------------------------------------------

//! == Operator: Testet auf Gleichheit zweier Vektoren
template < unsigned int Size, typename DataType >
bool SmallVector< Size, DataType >::operator==( const SmallVector< Size, DataType > & alius ) const
{
    double TOL = 1e-15;
    double diff = 0;
    for ( unsigned int i = 0; i < Size; ++i )
    {
        diff += Abs( (*this)[i] - alius[i] );
    }
    return ( diff < TOL );
}
// ------------------------------------------------------------------------------------------------

//! != Operator: Testet auf Ungleichheit zweier Vektoren
template < unsigned int Size, typename DataType >
bool SmallVector< Size, DataType >::operator!=( const SmallVector< Size, DataType > & alius ) const
{
    return !( *this == alius );
}
// ------------------------------------------------------------------------------------------------





//! Standard-Konstruktor
template < typename DataType >
SmallVector2< DataType >::SmallVector2()
: SmallVector< 2, DataType >()
{}

//! Konstruktor fuer 2 Elemente
template < typename DataType >
SmallVector2< DataType >::SmallVector2( DataType v1, DataType v2 )
{
    (*this)[0] = v1;
    (*this)[1] = v2;
}
// ------------------------------------------------------------------------------------------------

//! Konstruktor aus Elternklasse
template < typename DataType >
SmallVector2< DataType >::SmallVector2( const SmallVector<2, DataType> & v )
: SmallVector< 2, DataType > (v)
{}
// ------------------------------------------------------------------------------------------------





//! Standard-Konstruktor
template < typename DataType >
SmallVector3< DataType >::SmallVector3()
: SmallVector< 3, DataType >()
{}

//! Konstruktor fuer 3 Elemente
template < typename DataType >
SmallVector3< DataType >::SmallVector3( DataType v1, DataType v2, DataType v3 )
{
    (*this)[0] = v1;
    (*this)[1] = v2;
    (*this)[2] = v3;
}
// ------------------------------------------------------------------------------------------------

//! Konstruktor aus Elternklasse
template < typename DataType >
SmallVector3< DataType >::SmallVector3( const SmallVector<3, DataType> & v )
: SmallVector< 3, DataType > (v)
{}
// ------------------------------------------------------------------------------------------------




// explizte Instanzierung, damit das Kompilieren klappt
template class SmallVector< 2, double >;
template class SmallVector< 2, int >;
template class SmallVector< 3, double >;
template class SmallVector< 3, int >;
template class SmallVector< 4, int >;

template class SmallVector2< double >;
template class SmallVector2< int >;
template class SmallVector3< double >;
template class SmallVector3< int >;
