/** Klasse fuer die Dreiecke einer Triangulierung
 *
 *  \file   triangle.cpp
 *  \author Stefan W. von Deylen, Benedict Geihe
 *  \date   18.09.2008
 *
 */

#include "triangle.h"

#include <cmath>
#include <sstream>

using namespace std;

/**********************************************************************************************/
/*                                   class Triangle                                           */
/**********************************************************************************************/

//! Konstruktor aus 3 Knoten-Indices
Triangle::Triangle( GlobalIndex nodeIndex1, GlobalIndex nodeIndex2, GlobalIndex nodeIndex3 )
    : _AmTrans(2,3)
    , _index ( IndexNotSet )
    , _gridPtr ( NULL )
    , _markedForRefinement( false )
{
    setNodeIndex( 0, nodeIndex1 );
    setNodeIndex( 1, nodeIndex2 );
    setNodeIndex( 2, nodeIndex3 );

    setNeighbourIndex( 0, IndexNotSet );
    setNeighbourIndex( 1, IndexNotSet );
    setNeighbourIndex( 2, IndexNotSet );
}
// ------------------------------------------------------------------------------------------------

//! Lese globalen Index des Dreiecks
GlobalIndex Triangle::getIndex() const
{
    return _index;
}
// ------------------------------------------------------------------------------------------------

//! Lese globalen Index zu einem lokalen Knoten-Index
GlobalIndex Triangle::getNodeIndex( LocalIndex localNodeIndex ) const
{
    return _nodeIndex[localNodeIndex];
}
// ------------------------------------------------------------------------------------------------

//! Lese globalen Index des Nachbardreiecks (angegeben durch den lokalen Index des gegenueberliegenden Knotens)
GlobalIndex Triangle::getNeighbourIndex(LocalIndex localEdgeIndex) const
{
    return _neighbourIndex[localEdgeIndex];
}
// ------------------------------------------------------------------------------------------------

//! Setze globalen Dreiecksindex, gib alten zurueck
GlobalIndex Triangle::setIndex(GlobalIndex neuerIndex)
{
    GlobalIndex alterIndex = _index;
    _index = neuerIndex;
    return alterIndex;
}
// ------------------------------------------------------------------------------------------------

//! Setze globalen Knotenindex, gib alten zurueck
GlobalIndex Triangle::setNodeIndex( LocalIndex localNodeIndex, GlobalIndex globalNodeIndex )
{
    GlobalIndex alterNodeIndex = _nodeIndex[localNodeIndex];
    _nodeIndex[localNodeIndex] = globalNodeIndex;
    return alterNodeIndex;
}
// ------------------------------------------------------------------------------------------------

//! Setze globalen Nachbar-Dreiecksindex, gib alten zurueck
GlobalIndex Triangle::setNeighbourIndex( LocalIndex localEdgeIndex, GlobalIndex globalTriangleIndex )
{
    GlobalIndex alterEdgeIndex = _neighbourIndex[localEdgeIndex];
    _neighbourIndex[localEdgeIndex] = globalTriangleIndex;
    return alterEdgeIndex;
}
// ------------------------------------------------------------------------------------------------

//! Setze Zeiger auf das Gitter, zu dem dieses Dreieck gehoert
Grid * Triangle::setGrid(Grid * neuerGridPtr)
{
    Grid * alterGridPtr = _gridPtr;
    _gridPtr = neuerGridPtr;
    return alterGridPtr;
}
// ------------------------------------------------------------------------------------------------

//! Zugriff auf das Gitter
Grid & Triangle::getGrid()
{
    return *_gridPtr;
}
// ------------------------------------------------------------------------------------------------

//! Zugriff auf das Gitter (const)
const Grid & Triangle::getGrid() const
{
    return *_gridPtr;
}
// ------------------------------------------------------------------------------------------------

//! Zugriff auf Knoten
Node & Triangle::getNode(LocalIndex localNodeIndex)
{
    return getGrid().getNode(_nodeIndex[localNodeIndex]);
}
// ------------------------------------------------------------------------------------------------

//! Zugriff auf Knoten (const)
const Node & Triangle::getNode(LocalIndex localNodeIndex) const
{
    return getGrid().getNode(_nodeIndex[localNodeIndex]);
}
// ------------------------------------------------------------------------------------------------

//! Zugriff auf Nachbardreieck
Triangle & Triangle::getNeighbour(LocalIndex localEdgeIndex)
{
    return getGrid().getTriangle(_neighbourIndex[localEdgeIndex]);
}
// ------------------------------------------------------------------------------------------------

//! Zugriff auf Nachbardreieck (const)
const Triangle & Triangle::getNeighbour(LocalIndex localEdgeIndex) const
{
    return getGrid().getTriangle(_neighbourIndex[localEdgeIndex]);
}
// ------------------------------------------------------------------------------------------------

//! Gibt den Flaecheninhalt des Dreiecks zurueck
double Triangle::getVolume() const
{
    return _volume;
}
// ------------------------------------------------------------------------------------------------

//! Berechnet den Flaecheninhalt des Dreiecks
double Triangle::computeVolume() const
{
    const SmallVector<2> & v = _nodeDiff02;
    SmallVector<2> w = _nodeDiff12;

    // consider a triangle with vertices 0, v, w.
    // its volume computes as 1/2 * |v'| * |w|,
    // where v' is the projection of v onto w^\orthogonal,
    // which is v' = v - <v,w>/|w|^2 * w.

    double norm_w_sqr = w.norm_l2q();
    w *= -(v * w) / norm_w_sqr;
    w += v;
    // now w contains v'.

    return 0.5 * sqrt(norm_w_sqr) * w.norm_l2();
}
// ------------------------------------------------------------------------------------------------

//! Berechnet die Laenge einer Kante
double Triangle::getEdgeLength(LocalIndex localEdgeIndex) const
{
    if (localEdgeIndex == 1)
        return _nodeDiff02.norm_l2();
    if (localEdgeIndex == 0)
        return _nodeDiff12.norm_l2();

    SmallVector<2> d = _nodeDiff02;
    d -= _nodeDiff12;
    return d.norm_l2();
}
// ------------------------------------------------------------------------------------------------

//! Sucht die laengste Kante
LocalIndex Triangle::getLongestEdgeIndex() const
{
    double d0 = getEdgeLength(0),
           d1 = getEdgeLength(1),
           d2 = getEdgeLength(2);

    LocalIndex i = 0;
    if (d1 > d0)
    {
        // re-use d1 to store maximum length:
        d0 = d1;
        i = 1;
    }
    if (d2 > d0)
        i = 2;
    return i;
}
// ------------------------------------------------------------------------------------------------

//! Gibt die Laenge der laengsten Kanten aus
double Triangle::getLongestEdgeLength() const
{
    return getEdgeLength(getLongestEdgeIndex());
}
// ------------------------------------------------------------------------------------------------

//! Liefert den Mittelpunkt einer Kante
SmallVector2<double> Triangle::getEdgeMidpoint(LocalIndex localEdgeIndex) const
{
    SmallVector2<double> midpoint(getNode((localEdgeIndex + 1) % 3));
    midpoint *= 0.5;
    midpoint.addMultiple(0.5, getNode((localEdgeIndex + 2) % 3));
    return midpoint;
}
// ------------------------------------------------------------------------------------------------

//! Liefert den Schwerpunkt des Dreiecks
SmallVector2<double> Triangle::getCenter() const
{
    SmallVector2<double> midpoint(getNode(0));
    midpoint /= 3.;
    midpoint.addMultiple(1./3., getNode(1));
    midpoint.addMultiple(1./3., getNode(2));
    return midpoint;
}
// ------------------------------------------------------------------------------------------------

//! Berechnet normalisierten Aussen-Normalenvektor
SmallVector2<double> Triangle::getNormal(LocalIndex localEdgeIndex) const
{
    // Compute vector connecting nodes of specified edge
    SmallVector2<double> diff( getNode( (localEdgeIndex+1) % 3 ) );
    diff -= getNode( (localEdgeIndex+2) % 3 );

    // Rotate (clockwise)
    SmallVector2<double> normal( diff[1], -diff[0] );

    // Normalize
    normal.normalize();

    // Check Orientation
    diff  = getNode( localEdgeIndex );
    diff -= getNode( (localEdgeIndex+1) % 3 );
    if ( normal * diff > 0 ) normal *= -1.;

    return normal;
}
// ------------------------------------------------------------------------------------------------

//! Markiere fuer Verfeinerung
void Triangle::markForRefinement()
{
    _markedForRefinement = true;
}
// ------------------------------------------------------------------------------------------------

//! Loesche die Markierung
void Triangle::unsetRefinementMarker()
{
    _markedForRefinement = false;
}
// ------------------------------------------------------------------------------------------------

//! Prueft ob Dreieck markiert ist
bool Triangle::isMarkedForRefinement() const
{
    return _markedForRefinement;
}
// ------------------------------------------------------------------------------------------------

void Triangle::calcCartesianCoord ( const BarCoord & q, SmallVector2<> & cartCoord ) const
{
  cartCoord.setZero();
  for ( int i = 0; i < 3; ++i ) 
    cartCoord.addMultiple( q[i], getNode(i) );
}
// ------------------------------------------------------------------------------------------------
//! interpoliert f an einem Punkt q (übergeben in baryz. Koordinaten)
double Triangle::calcFunctionValue ( const Vector & f, const SmallVector3<> & q ) const {
   double ret = 0.;
   for ( int i = 0; i < 3; ++i )
     ret += q[i] * f[getNodeIndex(i)];

   return ret;
}

// ------------------------------------------------------------------------------------------------
//! Berechnet Gradienten in kartesischen Koordinaten
void Triangle::calcCartesianGrad ( const Vector & f, SmallVector2<> & gradF ) const
{
    // Berechne Gradienten in baryzentrischen Koordinaten
    SmallVector3<> barGrad ( f[getNodeIndex(0)], f[getNodeIndex(1)], f[getNodeIndex(2)] );

    // Wende Transformation an
    _AmTrans.apply( barGrad, static_cast<SmallVector2<> &>(gradF) );
}
// ------------------------------------------------------------------------------------------------

//! Berechnet Gradienten einer Basisfunktion in kartesischen Koordinaten auf dem aktuellen Dreieck 
void Triangle::calcCartesianBaseGrad ( LocalIndex index, SmallVector2<> & gradF ) const
{
    gradF[0] = _AmTrans(0,index);
    gradF[1] = _AmTrans(1,index);
}
// ------------------------------------------------------------------------------------------------

//! Funktion zur Berechnung diverser Hilfsvariablen
void Triangle::computeTransformation()
{
    Matrix _AInv(2,2);

    _nodeDiff02 = getNode(0);
    _nodeDiff02 -= getNode(2);
    _nodeDiff12 = getNode(1);
    _nodeDiff12 -= getNode(2);

    _AInv(0,0) = _nodeDiff02[0];
    _AInv(1,0) = _nodeDiff02[1];

    _AInv(0,1) = _nodeDiff12[0];
    _AInv(1,1) = _nodeDiff12[1];

    Matrix AT ( 2, 2);
    double detAInv = _AInv(0,0) * _AInv(1,1) - _AInv(0,1) * _AInv(1,0);
    AT(0,0) =   _AInv(1,1) / detAInv;
    AT(1,0) = - _AInv(0,1) / detAInv;
    AT(0,1) = - _AInv(1,0) / detAInv;
    AT(1,1) =   _AInv(0,0) / detAInv;

    Matrix TransCartToBarT(2,3);
    TransCartToBarT(0,0) = 1.;  TransCartToBarT(0,1) = 0.; TransCartToBarT(0,2) = -1.;
    TransCartToBarT(1,0) = 0.;  TransCartToBarT(1,1) = 1.; TransCartToBarT(1,2) = -1.;

    _AmTrans.mmult ( AT, TransCartToBarT );

    _volume = computeVolume();
}
// ------------------------------------------------------------------------------------------------

//! Ausgabefunktion mit <<-Operator
ostream & operator<< (ostream & str, const Triangle & triangle)
{
    str << "Index: ";
    if (triangle.getIndex() == IndexNotSet)
        str << "[not set]";
    else
        str << triangle.getIndex();

    str << endl
        << "Nodes: " << setw(3) << triangle.getNodeIndex(0) << "  " << triangle.getNode(0)
        << "       " << setw(3) << triangle.getNodeIndex(1) << "  " << triangle.getNode(1)
        << "       " << setw(3) << triangle.getNodeIndex(2) << "  " << triangle.getNode(2)
        << "Neighbours: " << triangle.getNeighbourIndex(0)
        << ", " << triangle.getNeighbourIndex(1)
        << ", " << triangle.getNeighbourIndex(2) << endl << endl;
    return str;
}
// ------------------------------------------------------------------------------------------------












/**********************************************************************************************/
/*                              class TriangleIteratorTemplate                                */
/**********************************************************************************************/

//! Kostruktor aus einem grid
template <typename TriangleType, typename GridType>
TriangleIteratorTemplate<TriangleType, GridType>::TriangleIteratorTemplate(GridType & grid)
    : _grid(grid)
    , _currentIndex(0)
{
}
// ------------------------------------------------------------------------------------------------

//! Iterator am Ende der Dreiecksliste?
template <typename TriangleType, typename GridType>
bool TriangleIteratorTemplate<TriangleType, GridType>::notAtEnd()
{
    return (_currentIndex < _grid._triangleList.size());
}
// ------------------------------------------------------------------------------------------------

//! Inkrement-Operator
template <typename TriangleType, typename GridType>
TriangleIteratorTemplate<TriangleType, GridType> & TriangleIteratorTemplate<TriangleType, GridType>::operator++()
{
    _currentIndex++;
    return *this;
}
// ------------------------------------------------------------------------------------------------

//! Dekrement-Operator
template <typename TriangleType, typename GridType>
TriangleIteratorTemplate<TriangleType, GridType> &
TriangleIteratorTemplate<TriangleType, GridType>::operator--()
{
    _currentIndex--;
    return *this;
}
// ------------------------------------------------------------------------------------------------

//! Dereferenzierung: Zugriff auf Triangle
template <typename TriangleType, typename GridType>
TriangleType & TriangleIteratorTemplate<TriangleType, GridType>::operator*()
{
    return _grid._triangleList[_currentIndex];
}
// ------------------------------------------------------------------------------------------------

//! Dereferenzierung: Zugriff auf member von Triangle
template <typename TriangleType, typename GridType>
TriangleType * TriangleIteratorTemplate<TriangleType, GridType>::operator->()
{
    return &(_grid._triangleList[_currentIndex]);
}
// ------------------------------------------------------------------------------------------------

//! Liefert globalen Index des aktuellen Dreiecks
template <typename TriangleType, typename GridType>
GlobalIndex TriangleIteratorTemplate<TriangleType, GridType>::getIndex() const
{
    return _currentIndex;
}
// ------------------------------------------------------------------------------------------------


// Explizite Instanzierung der verwendeten Iteratoren
template class TriangleIteratorTemplate<Triangle, Grid>;
template class TriangleIteratorTemplate<const Triangle, const Grid>;














/**********************************************************************************************/
/*                              class BoundaryIteratorTemplate                                */
/**********************************************************************************************/

//! Kostruktor aus einem grid
template <typename NodeType, typename GridType>
BoundaryIteratorTemplate<NodeType, GridType>::BoundaryIteratorTemplate(GridType & grid)
: _grid(grid)
, _dart(grid.getTriangle(0),0,0)
, _startIndex(0)
, _initialized(true)
{
  // Start looking for the boundary
  typename GridType::TriangleConstIterator tIter (_grid);
  int found = 0;
  while (( tIter.notAtEnd() ) && (!found))
  {
    for ( int i = 0; i < 3; ++i )
        if ( tIter->getNeighbourIndex(i) == IndexNotSet ) found = i+1;
    ++tIter;
  }
  if (!tIter.notAtEnd()) throw "BoundaryIterator: Could not find boundary edge!";
  --tIter;

  _startIndex = tIter->getNodeIndex(found%3);
  _dart.set( *tIter, found-1, found%3 );
}
// ------------------------------------------------------------------------------------------------

//! Iterator am Ende der Dreiecksliste?
template <typename NodeType, typename GridType>
bool BoundaryIteratorTemplate<NodeType, GridType>::notAtEnd() const
{
  return (( _initialized ) || ( _dart.getGlobalNodeIndex() != _startIndex ));
}
// ------------------------------------------------------------------------------------------------

//! Inkrement-Operator
template <typename NodeType, typename GridType>
BoundaryIteratorTemplate<NodeType, GridType> & BoundaryIteratorTemplate<NodeType, GridType>::operator++()
{
  _initialized = false;
  _dart.flipNode();
  _dart.flipEdge();
  while ( _dart.canFlipTriangle() )  {
    _dart.flipTriangle();
    _dart.flipEdge();
  }
  return *this;
}
// ------------------------------------------------------------------------------------------------

//! Dekrement-Operator
template <typename NodeType, typename GridType>
BoundaryIteratorTemplate<NodeType, GridType> & BoundaryIteratorTemplate<NodeType, GridType>::operator--()
{
  _initialized = false;
  _dart.flipEdge();
  while ( _dart.canFlipTriangle() )  {
    _dart.flipTriangle();
    _dart.flipEdge();
  }
  _dart.flipNode();
  return *this;
}
// ------------------------------------------------------------------------------------------------

//! Zugriff auf Knoten am Anfang des Segments
template <typename NodeType, typename GridType>
NodeType & BoundaryIteratorTemplate<NodeType, GridType>::getStart() const
{
  return _grid.getNode( _dart.getGlobalNodeIndex() );
}
// ------------------------------------------------------------------------------------------------

//! Zugriff auf Knoten am Ende des Segments
template <typename NodeType, typename GridType>
NodeType & BoundaryIteratorTemplate<NodeType, GridType>::getEnd() const
{
  return _grid.getNode( _dart.getNextNodeGlobalIndex() );
}
// ------------------------------------------------------------------------------------------------

//! Index des aktuellen Dreiecks
template <typename NodeType, typename GridType>
GlobalIndex BoundaryIteratorTemplate<NodeType, GridType>::getTriangleIndex() const
{
  return _dart.getGlobalTriangleIndex();
}
// ------------------------------------------------------------------------------------------------

//! Lokaler Index der aktuellen Kante
template <typename NodeType, typename GridType>
LocalIndex BoundaryIteratorTemplate<NodeType, GridType>::getEdgeIndex() const
{
  return _dart.getLocalEdgeIndex();
}
// ------------------------------------------------------------------------------------------------

//! Zugriff auf Indices der Knoten des Segments
template <typename NodeType, typename GridType>
void BoundaryIteratorTemplate<NodeType, GridType>::getIndices( SmallVector3<GlobalIndex> & ind ) const
{
  ind[0] = _dart.getGlobalNodeIndex();
  ind[1] = IndexNotSet;
  ind[2] = _dart.getNextNodeGlobalIndex();
}
// ------------------------------------------------------------------------------------------------

//! Knotenindex am Anfang des Segments
template <typename NodeType, typename GridType>
GlobalIndex BoundaryIteratorTemplate<NodeType, GridType>::getStartIndex() const
{
  return _dart.getGlobalNodeIndex();
}
// ------------------------------------------------------------------------------------------------

//! Knotenindex am Ende des Segments
template <typename NodeType, typename GridType>
GlobalIndex BoundaryIteratorTemplate<NodeType, GridType>::getEndIndex() const
{
  return _dart.getNextNodeGlobalIndex();
}
// ------------------------------------------------------------------------------------------------

//! Liefert Laenge des Segments
template <typename NodeType, typename GridType>
double BoundaryIteratorTemplate<NodeType, GridType>::getLength() const
{
  return _dart.getTriangle().getEdgeLength( _dart.getLocalEdgeIndex() );
}
// ------------------------------------------------------------------------------------------------


// Explizite Instanzierung der verwendeten Iteratoren
template class BoundaryIteratorTemplate<const Node, const Grid>;









/**********************************************************************************************/
/*                                  class DartIteratorTemplate                                */
/**********************************************************************************************/
//! Konstruktor aus Dreieck, Knoten, Kante
template <typename GridType>
DartIteratorTemplate<GridType>::
DartIteratorTemplate( typename DartIteratorTemplate<GridType>::TriangleType & triangle,
                      LocalIndex localEdgeIndex, LocalIndex localNodeIndex )
    : _grid (triangle.getGrid())
    , _triangle (triangle.getIndex())
    , _node(localNodeIndex)
    , _edge(localEdgeIndex)
{
}
// ------------------------------------------------------------------------------------------------

//! Konstruktor aus Dreieck und Kante
//! Der Knotenindex wird zu ( KantenIndex+1 ) mod 3 bestimmt
template <typename GridType>
DartIteratorTemplate<GridType>::
DartIteratorTemplate( typename DartIteratorTemplate<GridType>::TriangleType & triangle,
                      LocalIndex localEdgeIndex )
    : _grid (triangle.getGrid())
    , _triangle (triangle.getIndex())
    , _node((localEdgeIndex + 1) % 3)
    , _edge(localEdgeIndex)
{
}
// ------------------------------------------------------------------------------------------------

//! Modifikation
template <typename GridType>
void DartIteratorTemplate<GridType>::set( TriangleType & triangle, LocalIndex localEdgeIndex, LocalIndex localNodeIndex )
{
    if ( &_grid != &(triangle.getGrid()) ) throw "DartIteratorTemplate<GridType>::set: grids do not match!";
    _triangle = triangle.getIndex();
    _node = localNodeIndex;
    _edge = localEdgeIndex;
}
// ------------------------------------------------------------------------------------------------

//! Zugriff auf Dreieck
template <typename GridType>
typename DartIteratorTemplate<GridType>::TriangleType & DartIteratorTemplate<GridType>::getTriangle() const
{
    return _grid.getTriangle(_triangle);
}
// ------------------------------------------------------------------------------------------------

//! Zugriff auf Knoten
template <typename GridType>
typename DartIteratorTemplate<GridType>::NodeType & DartIteratorTemplate<GridType>::getNode() const
{
    return _grid.getTriangle(_triangle).getNode(_node);
}
// ------------------------------------------------------------------------------------------------

//! Liefert globalen Dreiecksindex
template <typename GridType>
GlobalIndex DartIteratorTemplate<GridType>::getGlobalTriangleIndex() const
{
    return _triangle;
}
// ------------------------------------------------------------------------------------------------

//! Liefert globalen Knotenindex
template <typename GridType>
GlobalIndex DartIteratorTemplate<GridType>::getGlobalNodeIndex() const
{
    return _grid.getTriangle(_triangle).getNodeIndex(_node);
}
// ------------------------------------------------------------------------------------------------

//! Liefert lokalen Knotenindex
template <typename GridType>
LocalIndex DartIteratorTemplate<GridType>::getLocalNodeIndex() const
{
    return _node;
}
// ------------------------------------------------------------------------------------------------

//! Liefert lokalen Kantenindex
template <typename GridType>
LocalIndex DartIteratorTemplate<GridType>::getLocalEdgeIndex() const
{
    return _edge;
}
// ------------------------------------------------------------------------------------------------

//! Liefert lokalen Index des anderen Knotens
template <typename GridType>
LocalIndex DartIteratorTemplate<GridType>::getNextNodeLocalIndex() const
{
    return 3 - (_node + _edge);
}
// ------------------------------------------------------------------------------------------------

//! Liefert lokalen Index der anderen Kante
template <typename GridType>
LocalIndex DartIteratorTemplate<GridType>::getNextEdgeLocalIndex() const
{
    return 3 - (_node + _edge);
}
// ------------------------------------------------------------------------------------------------

//! Liefert globalen Index des naechsten Knotens
template <typename GridType>
GlobalIndex DartIteratorTemplate<GridType>::getNextNodeGlobalIndex() const
{
    return _grid.getTriangle(_triangle).getNodeIndex(getNextNodeLocalIndex());
}
// ------------------------------------------------------------------------------------------------

//! Liefert globalen Index des auf der anderen Seite liegenden Dreiecks
template <typename GridType>
GlobalIndex DartIteratorTemplate<GridType>::getNextTriangleIndex() const
{
    return _grid.getTriangle(_triangle).getNeighbourIndex(_edge);
}
// ------------------------------------------------------------------------------------------------

//! Liefert lokalen Index des gemeinsamen Knotens
//! Prueft vorher ob gleiche Dreiecke und verschiedene Kanten uebergeben worden
template <typename GridType>
LocalIndex DartIteratorTemplate<GridType>::getCommonNodeLocalIndex(const DartIteratorTemplate<GridType> & d) const
{
    if ( getGlobalTriangleIndex() !=  d.getGlobalTriangleIndex() )
    {
        stringstream err_str;
        err_str << "DartIteratorTemplate<GridType>::getCommonNodeLocalIndex(): Passed DartIteratorTemplate<GridType> points "
                   "to triangle " << d.getGlobalTriangleIndex() << ", but this->triangle is "
                << getGlobalTriangleIndex() << ".";
        throw err_str.str();
    }
    if (getLocalEdgeIndex() == d.getLocalEdgeIndex())
    {
        stringstream err_str;
        err_str << "DartIteratorTemplate<GridType>::getCommonNodeLocalIndex(): Passed DartIteratorTemplate<GridType> and *this "
                   "point to the same edge with local GlobalIndex " << getLocalEdgeIndex() << ".";
        throw err_str.str();
    }
    return 3 - (getLocalEdgeIndex() + d.getLocalEdgeIndex());
}
// ------------------------------------------------------------------------------------------------

//! Liefert globalen Index des gemeinsamen Knotens
//! Dies ist auch moeglich (und gewuenscht) wenn verschiedene Dreiecke uebergeben worden
template <typename GridType>
GlobalIndex DartIteratorTemplate<GridType>::getCommonNodeGlobalIndex(const DartIteratorTemplate<GridType> & d) const
{
    if (   getGlobalNodeIndex() == d.getGlobalNodeIndex()
        || getGlobalNodeIndex() == d.getNextNodeGlobalIndex() )
        return getGlobalNodeIndex();

    if (   getNextNodeGlobalIndex() == d.getGlobalNodeIndex()
        || getNextNodeGlobalIndex() == d.getNextNodeGlobalIndex() )
        return getNextNodeGlobalIndex();

    stringstream err_str;
    err_str << "DartIteratorTemplate<GridType>::getCommonNodeGlobalIndex(): Passed Dart d ="
            << d << " and *this = " << *this << " have no node in common.";
    throw err_str.str();
}
// ------------------------------------------------------------------------------------------------

//! Prueft ob es ein Nachbardreieck gibt
template <typename GridType>
bool DartIteratorTemplate<GridType>::canFlipTriangle() const
{
    return (getNextTriangleIndex() != IndexNotSet);
}
// ------------------------------------------------------------------------------------------------

//! Wechselt zum Dreieck, das an der anderen Seite der aktuellen Kante liegt
template <typename GridType>
void DartIteratorTemplate<GridType>::flipTriangle()
{
    GlobalIndex newTriangleIndex = getNextTriangleIndex();
    const Triangle & newTriangle = _grid.getTriangle(newTriangleIndex);
    GlobalIndex globalNodeIndex = getTriangle().getNodeIndex(_node);
    LocalIndex otherNodesLocalIndex = 3 - (_node + _edge);
    GlobalIndex globalOtherNodesIndex = getTriangle().getNodeIndex(otherNodesLocalIndex);

    // find out which node of the new triangle is ours
    LocalIndex newNodeIndex = IndexNotSet,
        newOtherNodesIndex = IndexNotSet;
    for (LocalIndex i = 0; i < 3; ++i)
        if (newTriangle.getNodeIndex(i) == globalNodeIndex)
        {
            newNodeIndex = i;
            break;
        }
    // find out which node of the new triangle lies on our edge
    for (LocalIndex i = 1; i < 3; ++i)
        if (newTriangle.getNodeIndex((newNodeIndex + i) % 3) == globalOtherNodesIndex)
        {
            newOtherNodesIndex = (newNodeIndex + i) % 3;
            break;
        }

    if (newNodeIndex == IndexNotSet || newOtherNodesIndex == IndexNotSet)
    {
        stringstream err_msg;
        err_msg << "DartIteratorTemplate<GridType>::flipTriangle(): found inconsistency in grid: triangle "
                << newTriangle << " should be neighbour of " << getTriangle()
                << ", but does not contain node " << _grid.getNode(globalNodeIndex) << " or "
                << _grid.getNode(globalOtherNodesIndex) << ".";
        throw err_msg.str();
    }

    LocalIndex newEdgeIndex = (-(newNodeIndex + newOtherNodesIndex)) % 3;
    if (newEdgeIndex < 0) newEdgeIndex += 3;

    // all information found, now set:
    _triangle = newTriangleIndex;
    _node = newNodeIndex;
    _edge = newEdgeIndex;
}
// ------------------------------------------------------------------------------------------------

//! Wechselt zum Knoten, der an der anderen Seite der aktuellen Kante liegt
template <typename GridType>
void DartIteratorTemplate<GridType>::flipNode()
{
    _node = getNextNodeLocalIndex();
}
// ------------------------------------------------------------------------------------------------

//! Wechselt zur Kante, die im selben Dreieck an der anderen Seite des aktuellen Knotens liegt
template <typename GridType>
void DartIteratorTemplate<GridType>::flipEdge()
{
    _edge = getNextEdgeLocalIndex();
}
// ------------------------------------------------------------------------------------------------

//! Ausgabefunktion mit <<-Operator
template <typename GridType>
ostream & operator<< (ostream & str, const DartIteratorTemplate<GridType> & d)
{
    str << "[triangle = " << d.getGlobalTriangleIndex() << ", edge = " << d.getLocalEdgeIndex()
        << ", node = " << d.getLocalNodeIndex() << "]";
    return str;
}
// ------------------------------------------------------------------------------------------------

// explizite Instanzierung
template class DartIteratorTemplate<Grid>;
template class DartIteratorTemplate<const Grid>;

template ostream & operator<< (ostream &, const DartIteratorTemplate<Grid> &);
template ostream & operator<< (ostream &, const DartIteratorTemplate<const Grid> &);
