/** Klasse fuer die Dreiecke einer Triangulierung
 *
 *  \file   triangle.h
 *  \author Stefan W. von Deylen, Benedict Geihe
 *  \date   18.09.2008
 *
 */

#ifndef __triangle_h
#define __triangle_h

#include <iostream>
#include <vector>

using namespace std;

class Triangle;
class Grid;

typedef vector<Triangle> TriangleList;
template <typename TriangleType, typename GridType> class TriangleIteratorTemplate;

template <typename NodeType, typename GridType> class BoundaryIteratorTemplate;

template <typename GridType> class DartIteratorTemplate;


#include "smallVector.h"
#include "grid.h"
#include "../fe/quadrature.h"
#include "../vecmat/vector.h"
#include "../vecmat/matrix.h"

//! Dreiecks-Klasse
class Triangle
{
    public:
        // Konstruktor
        Triangle( GlobalIndex nodeIndex1, GlobalIndex nodeIndex2, GlobalIndex nodeIndex3 );

        // Index-Getter
        GlobalIndex getIndex() const;
        GlobalIndex getNodeIndex(LocalIndex localNodeIndex) const;
        GlobalIndex getNeighbourIndex(LocalIndex localEdgeIndex) const;

        // Zugriffe auf Gitter, Knoten des Dreiecks, Nachbardreiecke
              Grid & getGrid();
        const Grid & getGrid() const;
              Node & getNode(LocalIndex localNodeIndex);
        const Node & getNode(LocalIndex localNodeIndex) const;
              Triangle & getNeighbour(LocalIndex localEdgeIndex);
        const Triangle & getNeighbour(LocalIndex localEdgeIndex) const;

        // geometrische Funktionen (Volumen, Schwerpunkt etc.)
        double getVolume() const;
        double getEdgeLength(LocalIndex localEdgeIndex) const;
        double getLongestEdgeLength() const;
        LocalIndex getLongestEdgeIndex() const;
        SmallVector2<double> getEdgeMidpoint(LocalIndex localEdgeIndex) const;
        SmallVector2<double> getCenter() const;
        SmallVector2<double> getNormal(LocalIndex localEdgeIndex) const;

        // Verfeinerung
        void markForRefinement();
        void unsetRefinementMarker();
        bool isMarkedForRefinement() const;

        void calcCartesianCoord ( const BarCoord & q, SmallVector2<> & cartCoord ) const;
        double calcFunctionValue ( const Vector & f, const SmallVector3<> & barCoord ) const;
        // Berechnung von Gradienten in kartesischen Koordinaten
        void calcCartesianGrad ( const Vector & f, SmallVector2<> & gradF ) const;
        void calcCartesianBaseGrad ( LocalIndex index, SmallVector2<> & gradF ) const;

    protected:
        // Methoden zum Setzen der Daten
        GlobalIndex setIndex(GlobalIndex neuerIndex);
        GlobalIndex setNodeIndex(int localNodeIndex, GlobalIndex globalNodeIndex);
        GlobalIndex setNeighbourIndex(int localEdgeIndex, GlobalIndex globalTriangleIndex);
        Grid * setGrid(Grid * neuerGridPtr);

        // Hilfsfunktion und -variablen zur schnelleren
        // Basisfunktions-Auswertung
        void computeTransformation();
        double computeVolume() const;

        Matrix _AmTrans;
        double _volume;
        SmallVector2<> _nodeDiff02;
        SmallVector2<> _nodeDiff12;


        // Index-Informationen ueber Knoten und Nachbarn
        GlobalIndex _nodeIndex[3];
        GlobalIndex _neighbourIndex[3];

        // Informationen ueber eigenen Index, Gitter und Markierung zur Verfeinerung
        GlobalIndex _index;
        Grid *      _gridPtr;
        bool        _markedForRefinement;

        // Gestatte der Gitter-Klasse und der Steifigkeits-Bilinearform Zugriff auf die geschuetzten Funktionen und Daten
        friend class Grid;
        friend class StiffBilf;
};

//! Ausgabefunktion mit <<-Operator
ostream & operator<< ( ostream & str, const Triangle & triangle );


// ----------------------------------------------------------------------------------------------------------------------


//! Iterator ueber alle Dreiecke
//! (Template fuer die const und nicht-const Variante)
template <typename TriangleType, typename GridType>
class TriangleIteratorTemplate
{
    public:
        TriangleIteratorTemplate(GridType & grid);

        bool notAtEnd();
        TriangleIteratorTemplate & operator++();
        TriangleIteratorTemplate & operator--();

        TriangleType & operator*();
        TriangleType * operator->();

        GlobalIndex getIndex() const;

    protected:
        GridType & _grid;
        size_t _currentIndex;
};


// ----------------------------------------------------------------------------------------------------------------------


//! Dart-Iterator
//! (Template fuer die const und nicht-const Variante)
template <typename GridType>
class DartIteratorTemplate
{
    public:
        typedef typename TypeMapper<GridType>::TriangleType TriangleType;
        typedef typename TypeMapper<GridType>::NodeType NodeType;

        DartIteratorTemplate( TriangleType & triangle, LocalIndex localEdgeIndex, LocalIndex localNodeIndex );
        DartIteratorTemplate( TriangleType & triangle, LocalIndex localEdgeIndex );

        // Modifikation
        void set( TriangleType & triangle, LocalIndex localEdgeIndex, LocalIndex localNodeIndex );

        // Zugriff auf Knoten, Dreieck
        TriangleType & getTriangle() const;
        NodeType & getNode() const;

        // Zugriff auf globale Indices
        GlobalIndex getGlobalTriangleIndex() const;
        GlobalIndex getGlobalNodeIndex() const;

        // Zugriff auf lokale Indices
        LocalIndex getLocalNodeIndex() const;
        LocalIndex getLocalEdgeIndex() const;

        // Zugriff auf Indices des naechsten Knoten, der naechsten Kante, des naechsten Dreiecks
        LocalIndex  getNextNodeLocalIndex() const;
        LocalIndex  getNextEdgeLocalIndex() const;
        GlobalIndex getNextNodeGlobalIndex() const;
        GlobalIndex getNextTriangleIndex() const;

        // Zugriff auf Index eines gemeinsamen Knotens
        LocalIndex  getCommonNodeLocalIndex ( const DartIteratorTemplate & d ) const;
        GlobalIndex getCommonNodeGlobalIndex( const DartIteratorTemplate & d ) const;

        // Bewegung auf dem Gitter
        bool canFlipTriangle() const;
        void flipTriangle();
        void flipNode();
        void flipEdge();

    protected:
        GridType & _grid;
        GlobalIndex _triangle;
        LocalIndex  _node;
        LocalIndex  _edge;
};

// ----------------------------------------------------------------------------------------------------------------------

template <typename GridType>
ostream & operator<< (ostream & str, const DartIteratorTemplate<GridType> & d);

// ----------------------------------------------------------------------------------------------------------------------


//! Iterator ueber alle Randsegmente
//! (Template fuer die const und nicht-const Variante)
template <typename NodeType, typename GridType>
class BoundaryIteratorTemplate
{
public:
  BoundaryIteratorTemplate(GridType & grid);

  bool notAtEnd() const;
  BoundaryIteratorTemplate & operator++();
  BoundaryIteratorTemplate & operator--();

  NodeType & getStart() const;
  NodeType & getEnd() const;
  GlobalIndex getTriangleIndex() const;
  LocalIndex getEdgeIndex() const;
  void getIndices( SmallVector3<GlobalIndex> & ind ) const;
  GlobalIndex getStartIndex() const;
  GlobalIndex getEndIndex() const;
  double getLength() const;

protected:
  GridType & _grid;
  typename GridType::DartConstIterator _dart;
  GlobalIndex _startIndex;
  bool _initialized;
};

// ----------------------------------------------------------------------------------------------------------------------

#endif
