/** Matrix-Klasse für das Programmierpraktikum
 *
 *  \file   matrix.cpp
 *  \author Stefan W. von Deylen, Benedict Geihe
 *  \date   18.09.2008
 *
 */

#include "matrix.h"


using namespace std;

//! Konstruktor
Matrix::Matrix()
: _n( 0 ), _m( 0 )
{}
// ------------------------------------------------------------------------------------------------

//! Konstruktor mit Dimensionsangaben
Matrix::Matrix( unsigned int n, unsigned int m)
: Vector( n * m ), _n( n ), _m( m )
{}
// ------------------------------------------------------------------------------------------------

//! Kopier-Konstruktor
Matrix::Matrix( const Matrix& mat )
: Vector( mat ), _n( mat.getRows() ), _m( mat.getCols() )
{}
// ------------------------------------------------------------------------------------------------

//! Destruktor
Matrix::~Matrix()
{
    // Der Destruktor der Basisklasse (Vector) wird automatisch aufgerufen
}
// ------------------------------------------------------------------------------------------------

//! Zuweisungsoperator
Matrix& Matrix::operator=( const Matrix& mat )
{
    // Selbst-Zuweisung abfangen
    if ( this == &mat ) return *this;
    // Ansonsten
    (*this).vector<double>::operator=( mat );
    _n = mat.getRows();
    _m = mat.getCols();
    return *this;
}
// ------------------------------------------------------------------------------------------------

//! Dimensionsausgabe Zeilen
unsigned int Matrix::getRows() const
{
    return _n;
}
// ------------------------------------------------------------------------------------------------

//! Dimensionsausgabe Zeilen
unsigned int Matrix::getCols() const
{
    return _m;
}
// ------------------------------------------------------------------------------------------------

//! Elementzugriff lesend
double Matrix::operator()( unsigned int i, unsigned int j ) const
{
    assert( i < _n );
    assert( j < _m );

    return (*this)[ _m * i + j ];
}
// ------------------------------------------------------------------------------------------------

//! Elementzugriff schreibend
double& Matrix::operator()( unsigned int i, unsigned int j )
{
    assert( i < _n );
    assert( j < _m );

    return (*this)[ _m * i + j ];
}
// ------------------------------------------------------------------------------------------------

//! Matrix-Multiplikation: *this = A * B (A,B Matrizen)
void Matrix::mmult(const Matrix& A, const Matrix& B)
{
    Vector::setZero();
    assert(          _n == A.getRows() );
    assert(          _m == B.getCols() );
    assert( A.getCols() == B.getRows() );

    for ( unsigned int i = 0; i < _n; i++ )
    {
        for ( unsigned int j = 0; j < _m; j++ )
        {
            for ( unsigned int k = 0; k < A.getCols(); k++ )
            {
                (*this)(i,j) += A(i,k) * B(k,j);
            }
        }
    }
}
// ------------------------------------------------------------------------------------------------

//! Transposition (in-place)
void Matrix::transpose()
{
    assert ( _n == _m );

    for (unsigned int i = 0; i < _n; ++i)
    {
        for (unsigned int j = 0; j < _m; ++j)
        {
            std::swap ( (*this)(i,j), (*this)(j,i) );
        }
    }
}
// ------------------------------------------------------------------------------------------------

//! Matrix-Vektor-Multiplikation: b = *this * x * lambda berechnen
//! \param x Vektor (Argument)
//! \param b reeller Faktor (Standardwert 1.0)
void Matrix::apply( const Vector& x, Vector& b, double lambda ) const
{
    assert( x.size() == _m );
    assert( b.size() == _n );

    b.setZero();
    for ( unsigned int i = 0; i < _n; i++ )
    {
        for ( unsigned int j = 0; j < _m; j++ )
        {
            b[i] += (*this)(i,j) * x[j];
        }
        b[i] *= lambda;
    }
}
// ------------------------------------------------------------------------------------------------

//! Matrix-Vektor-Multiplikation: b += *this * x * lambda berechnen
//! \param x Vektor (Argument)
//! \param b reeller Faktor (Standardwert 1.0)
void Matrix::applyAdd( const Vector& x, Vector& b, double lambda ) const
{
    assert( x.size() == _m );
    assert( b.size() == _n );

    for ( unsigned int i = 0; i < _n; i++ )
    {
        for ( unsigned int j = 0; j < _m; j++ )
        {
            b[i] += (*this)(i,j) * x[j] * lambda;
        }
    }
}
// ------------------------------------------------------------------------------------------------

//! l1-Norm, d. h. Spaltensummennorm
double Matrix::norm_l1() const
{
    double max = 0.0;
    for ( unsigned int j = 0; j < _m; ++j)
    {
        double s = 0.0;
        for ( unsigned int i = 0; i < _n; ++i )
        {
            s += fabs( (*this)(i,j) );
        }
        if ( s > max ) max = s;
    }
    return max;
}
// ------------------------------------------------------------------------------------------------

//! l-unendlich-Norm, d. h. Zeilensummennorm
double Matrix::norm_linf() const
{
    double max = 0;
    for ( unsigned int i = 0; i < _n; i++ )
    {
        double s = 0;
        for ( unsigned int j = 0; j < _m; j++ )
        {
            s += fabs( (*this)(i,j) );
        }
        if ( s > max ) max = s;
    }
    return max;
}
// ------------------------------------------------------------------------------------------------

//! Ausgabefunktion mit <<-Operator
ostream& operator<<( ostream& stream, const Matrix& matrix )
{
    stream.precision(2);
    stream.setf( ios::fixed, ios::floatfield );

    if ( matrix.getRows() * matrix.getCols() > 0 )
    {
        for ( unsigned int i = 0; i < matrix.getRows(); i++ )
        {
            stream << endl << "|";
            for ( unsigned int j = 0; j < matrix.getCols(); j++ )
            {
                stream << " " << setw(8) << matrix(i,j) << " ";
            }
            stream << "|";
        }
    }
    else stream << "| |";
    stream << endl << endl;

    stream.precision(6);
    stream.unsetf( ios::fixed );

    return stream;
}
// ------------------------------------------------------------------------------------------------
