/** Löser-Klasse für das Programmierpraktikum
 *
 *  \file   solver.cpp
 *  \author Stefan W. von Deylen, Benedict Geihe
 *  \date   23.09.2008
 *
 */

#include "solver.h"
#include "matrix.h"     // fuer explizite Instanzierung, s.u.
#include "sparse.h"

using namespace std;

//! Konstruktor
//! \param matrix   Matrix des zu loesenden LGS
//! \param verbose  Schaltet ausfuehrliche Ausgaben ein/aus
template <typename MatrixType>
Solver< MatrixType >::Solver( const MatrixType & matrix, bool verbose )
: _matrix( matrix ), _verbose( verbose ), _tol( 1e-10),
  _error( 0.0 ), _firsterror( 0.0 ), _lasterror( 0.0 ), _constrViolation( 0.0 ), _projectionThreshold( 1e-12 ),
  _maxiter( matrix.getRows() * 10 ),
  _constr(NULL)
{}


//! Schaltet ausfuehrliche Ausgaben ein/aus
template <typename MatrixType>
void Solver< MatrixType >::setVerbose( bool verbose )
{
    _verbose = verbose;
}

//! Setzt Abbruch-Genauigkeit fuer iterative Verfahren
template <typename MatrixType>
void Solver< MatrixType >::setTol( double tol )
{
    _tol = tol;
}

//! Setzt maximale Iterationszahl fuer iterative Verfahren
template <typename MatrixType>
void Solver< MatrixType >::setMaxIter( int maxiter )
{
    _maxiter = maxiter;
}

//! Sets threshold for performing projection when using projectingCGapply
template <typename MatrixType>
void Solver< MatrixType >::setProjectionThreshold( double threshold )
{
    _projectionThreshold = threshold;
}



//! CG-Verfahren
//! \param   b  rechte Seite
//! \param   x  Loesung ( der Inhalt wird als Startwert verwendet und anschließend ueberschrieben! )
//! \warning Die Matrix muß symmetrisch sein!
template <typename MatrixType>
void Solver< MatrixType >::CGapply( const Vector & b, Vector & x )
{
    Vector r( x.size() );     // Residuum
    Vector p( x.size() );     // Suchrichtung
    Vector tmp( x.size() );   // A * p
    double alpha,             // Schrittweite
           beta,              // Zur Bestimmung der neuen Suchrichtung
           norm_rq,           // Norm^2(r_k)
           norm_rq_alt;       // Norm^2(r_{k-1})

    // determine initial constraint violation
    if( _constr )
      _constrViolation = (*_constr) * x;
  
    // r_0 = b - A * x
    _matrix.apply( x, r, -1.0 );
    r.addMultiple( 1.0, b );

    // Erste Suchrichtung p = r_0
    p = r;

    norm_rq = r.norm_l2q();
    _error = _firsterror = _lasterror = sqrt( norm_rq ); _iter = 1;
    if ( _verbose) printHeader();

    while ( ( _iter < _maxiter) && ( _error > _tol ) )
    {
        _matrix.apply( p, tmp );             // tmp = A * p
        alpha = ( r * p ) / ( p * tmp );     // alpha = r * p / (p^T * A * p)
        x.addMultiple( alpha, p );           // x = x + alpha * p
        r.addMultiple( -alpha, tmp );        // r = r - alpha * A * p

        // new constraint violation
        if( _constr )
        {
          _constrViolation = (*_constr) * x;
          if ( abs(_constrViolation) > _projectionThreshold )
          {
            x.addMultiple ( -_constrViolation, *_constr );
            _constrViolation = (*_constr) * x;
          }
        }
    
        norm_rq_alt = norm_rq;
        norm_rq     = r.norm_l2q();

        // Berechnung der Suchrichtung für den nächsten Schritt
        beta = norm_rq / norm_rq_alt;
        // p_neu = beta * p_alt + r
        p *= beta;
        p += r;

        _error = sqrt( norm_rq );
        if ( _verbose ) printIterate();
        _lasterror = _error;
        _iter++;
    }
    if ( _error > _tol ) cerr << endl << "!!! Could not solve to required tolerance ( "
                              << _tol << " ) within " << _maxiter << " Iterations !!!" << endl << endl;
}

//! CG-Verfahren
//! \param   b  rechte Seite
//! \param   x  Loesung ( wird zu bisherigem Inhalt addiert! )
//! \warning Die Matrix muß symmetrisch sein!
template <typename MatrixType>
void Solver< MatrixType >::CGapplyAdd( const Vector & b, Vector & x )
{
    Vector save( x );
    x.setZero();
    CGapply( b, x );
    x += save;
}


//! CG method with projection to keep equality constraint \f$ constr \cdot x = 0 \f$
//! \param   constr  vector for constraint (normalized!)
//! \param   b       right hand side
//! \param   x       solution ( will be added to current vector! )
//! \warning Matrix has to be symmetric!
template <typename MatrixType>
void Solver< MatrixType >::projectingCGapply( const Vector & constr, const Vector & b, Vector & x )
{
    // set constraint
    if ( _constr )
      delete _constr;
    _constr = new Vector( constr );

    // apply CG solver
    //! \bug Currently the solution depends on the initail data, hence we set this to 0! 
    x.setZero();
    CGapply( b, x );

    // remove constraint
    delete _constr;
    _constr = NULL;

}

//! CG method with projection to keep equality constraint \f$ constr \cdot x = 0 \f$
//! \param   constr  vector for constraint (normalized!)
//! \param   b       right hand side
//! \param   x       solution ( will be added to current vector! )
//! \warning Matrix has to be symmetric!
template <typename MatrixType>
void Solver< MatrixType >::projectingCGapplyAdd( const Vector & constr, const Vector & b, Vector & x )
{
    Vector save( x );
    x.setZero();
    projectingCGapply( constr, b, x );
    x += save;
}



// private Hilfsfunktionen
template <typename MatrixType>
void Solver< MatrixType >::printHeader() const
{
    // checking _constrViolation == 0.0 is intended and will most likely just work
    cout << "Iteration:    " << setw(20) << left << "Residual(l2):" << setw(20)
         << "Reduction (step)" << setw(20) << "Reduction (total)";
    if ( _constrViolation != 0.0 ) cout << setw(20) << "Constraint violation";
    cout << endl;
    cout << "-----------------------------------------------------------------------";
    if ( _constrViolation != 0.0 ) cout << "----------------------";
    cout << endl;
    cout << setw(10) << right << "0"  << "    " << setw(20) << left << _error
         << setw(20) << "-" << setw(20) << "-";
    if ( _constrViolation != 0.0 ) cout << setw(20) << _constrViolation;
    cout << endl;
}

template <typename MatrixType>
void Solver< MatrixType >::printIterate() const
{
    cout << setw(10) << right << _iter << "    "
         << setw(20) << left << _error
         << setw(20) << _lasterror  / _error
         << setw(20) << _firsterror / _error;
    if ( _constrViolation != 0.0 ) cout << setw(20) << _constrViolation;
    cout << endl;
}


// explizte Instanzierung, damit das Kompilieren klappt!
template class Solver< Matrix >;
template class Solver< SparseMatrix >;
