/** Duennbesetzte Matrix-Klasse für das Programmierpraktikum
 *
 *  \file   sparse.cpp
 *  \author Stefan W. von Deylen, Benedict Geihe
 *  \date   24.09.2008
 *
 */

#include "sparse.h"

using namespace std;

//! Konstruktor
SparseMatrix::SparseMatrix()
: _n( 0 ), _m( 0 )
{}

//! Konstruktor mit Dimensionsangaben
SparseMatrix::SparseMatrix( unsigned int n, unsigned int m)
: vector<SparseRow>( n ), _n( n ), _m( m )
{}

//! Kopier-Konstruktor
SparseMatrix::SparseMatrix( const SparseMatrix& mat )
: vector<SparseRow>( mat ), _n( mat.getRows() ), _m( mat.getCols() )
{}

//! Destruktor
SparseMatrix::~SparseMatrix()
{
    // Der Destruktor der Basisklasse (vector) wird automatisch aufgerufen
    // Elementare Membervariablen werden automatisch zerstoert
}

//! Dimensionsausgabe Zeilen
unsigned int SparseMatrix::getRows() const
{
    return _n;
}

//! Dimensionsausgabe Zeilen
unsigned int SparseMatrix::getCols() const
{
    return _m;
}

//! Elementzugriff lesend
double SparseMatrix::operator()( unsigned int i, unsigned int j ) const
{
    assert( i < _n );
    assert( j < _m );

    // Falls an der Stelle i,j kein Eintrag steht, soll auch keiner durch den []-Operator erstellt werden!
    SparseRow::const_iterator it = (*this)[i].find(j);
    if ( it != (*this)[i].end() ) return it->second;
    else return 0.0;
}

//! Elementzugriff schreibend
double& SparseMatrix::operator()( unsigned int i, unsigned int j )
{
    assert( i < _n );
    assert( j < _m );

    // Hier ok. Da sowieso geschrieben wird, kann/muss der Eintrag auch neu erstellt werden!
    return (*this)[i][j];
}

//! Setze alle Werte auf 0, d.h. leere alle maps
void SparseMatrix::setZero()
{
    for ( unsigned int i = 0; i < _n; i++ )  {
        (*this)[i].clear();
    }
}

//! Setze Wert an (i,j) auf 0, d.h. lösche entsprechenden Eintrag aus entsprechender map
void SparseMatrix::setZero( unsigned int i, unsigned int j )
{
    (*this)[i].erase(j);
}

//! *this = d * mat setzen (d reell, mat Matrix)
void SparseMatrix::setToMultiple( const double d, const SparseMatrix& mat )
{
    assert( _n == mat.getRows() );
    assert( _m == mat.getCols() );

    SparseRow::const_iterator iter;
    for ( unsigned int i = 0; i < _n; i++ )  {
        (*this)[i].clear();
        for ( iter = mat[i].begin() ; iter != mat[i].end(); ++iter )  {
            (*this)[i][iter->first] = iter->second * d;
        }
    }
}

//! *this += d * v berechnen (d reell, mat Matrix)
void SparseMatrix::addMultiple( const double d, const SparseMatrix& mat )
{
    assert( _n == mat.getRows() );
    assert( _m == mat.getCols() );

    SparseRow::const_iterator iter;
    for ( unsigned int i = 0; i < _n; i++ )  {
        for ( iter = mat[i].begin() ; iter != mat[i].end(); ++iter )  {
            (*this)[i][iter->first] += iter->second * d;
        }
    }
}

//! Matrix-Multiplikation: *this = A * B (A,B Matrizen)
void SparseMatrix::mmult(const SparseMatrix& A, const SparseMatrix& B)
{
    assert(          _n == A.getRows() );
    assert(          _m == B.getCols() );
    assert( A.getCols() == B.getRows() );

    double tmp;
    SparseRow::const_iterator iter;
    for ( unsigned int i = 0; i < _n; i++ )  {
        (*this)[i].clear();                         // Bisherige Matrix wird ueberschrieben
        for ( unsigned int j = 0; j < _m; j++ )  {
            tmp = 0.0;
            for ( iter = A[i].begin() ; iter != A[i].end(); ++iter )  {
                tmp+= iter->second * B(iter->first,j);
            }
            if ( fabs(tmp) > 1e-15 ) (*this)[i][j] = tmp;
        }
    }
}


//! \brief Loescht angegebene Zeile und Spalte
//! und setzt Diagonaleintrag dort auf \a diagEntry.
void SparseMatrix::setRowColToDiagonal ( const int index, double diagEntry )
{
    SparseRow::const_iterator iter;
    for ( size_t i = 0; i < _n; i++ )
	setZero( i, index );
    (*this)[index].clear();
    (*this)(index, index) = diagEntry;
}


//! \brief Addiert Vielfaches einer Spalte zu einer anderen
//! \param d       Faktor fuer Vielfaches
//! \param fromCol Spalte, die addiert werden soll
//! \param toCol   Spalte, zu der addiert werden soll
void SparseMatrix::addMultipleCol( const double d, const int fromCol, const int toCol )
{
    SparseRow::const_iterator iter;
    for ( unsigned int i = 0; i < _n; ++i )  {
        iter = (*this)[i].find(fromCol);
	if ( iter != (*this)[i].end() ) (*this)[i][toCol] += d * iter->second;
    }
}


//! \brief Addiert Vielfaches einer Zeile zu einer anderen
//! \param d       Faktor fuer Vielfaches
//! \param fromRow Zeile, die addiert werden soll
//! \param toRow   Zeile, zu der addiert werden soll
void SparseMatrix::addMultipleRow( const double d, const int fromRow, const int toRow )
{
    SparseRow::const_iterator iter;
    for ( iter = (*this)[fromRow].begin() ; iter != (*this)[fromRow].end(); ++iter )
	(*this)[toRow][iter->first] += d * iter->second;
}


//! Matrix-Vektor-Multiplikation: b = *this * x * lambda berechnen
//! \param x Vektor (Argument)
//! \param b reeller Faktor (Standardwert 1.0)
void SparseMatrix::apply( const Vector& x, Vector& b, double lambda ) const
{
    b.setZero();
    applyAdd(x, b, lambda);
}

//! Matrix-Vektor-Multiplikation: b += *this * x * lambda berechnen
//! \param x Argument-Vektor
//! \param b Ziel-Vektor
//! \param lambda reeller Faktor (Standardwert 1.0)
void SparseMatrix::applyAdd( const Vector& x, Vector& b, double lambda ) const
{
    assert( x.size() == _m );
    assert( b.size() == _n );

    SparseRow::const_iterator iter;
    for ( unsigned int i = 0; i < _n; i++ )  {
        for ( iter = (*this)[i].begin(); iter != (*this)[i].end(); ++iter )  {
            b[i] += iter->second * x[iter->first] * lambda;
        }
    }
}

//! l1-Norm, d. h. Spaltensummennorm
//! Leider nicht effizienter als bei vollbesetzten Matrizen...
double SparseMatrix::norm_l1() const
{
    double max = 0.0;
    for ( unsigned int j = 0; j < _m; ++j)
    {
        double s = 0.0;
        for ( unsigned int i = 0; i < _n; ++i )  {
            s += fabs( (*this)(i,j) );
        }
        if ( s > max ) max = s;
    }
    return max;
}

//! l-unendlich-Norm, d. h. Zeilensummennorm
double SparseMatrix::norm_linf() const
{
    SparseRow::const_iterator iter;
    double max = 0;
    for ( unsigned int i = 0; i < _n; i++ )
    {
        double s = 0;
        for ( iter = (*this)[i].begin(); iter != (*this)[i].end(); ++iter )  {
            s += fabs( iter->second );
        }
        if ( s > max ) max = s;
    }
    return max;
}

//! Ausgabefunktion mit <<-Operator
ostream& operator<<( ostream& stream, const SparseMatrix& matrix )
{
    stream.precision(2);
    stream.setf( ios::fixed, ios::floatfield );

    if ( matrix.getRows() * matrix.getCols() > 0 )
    {
        for ( unsigned int i = 0; i < matrix.getRows(); i++ )  {
            stream << endl << "|";
            for ( unsigned int j = 0; j < matrix.getCols(); j++ )  {
                stream << " " << setw(8) << matrix(i,j) << " ";
            }
            stream << "|";
        }
    }
    else stream << "| |";
    stream << endl << endl;

    stream.precision(6);
    stream.unsetf( ios::fixed );

    return stream;
}
