/** Vektor-Klasse für das Programmierpraktikum
 *
 *  \file   vector.cpp
 *  \author Stefan W. von Deylen, Benedict Geihe
 *  \date   16.09.2008
 *
 */

#include "vector.h"
#include "../triang/grid.h"


using namespace std;


//! Konstruktor
Vector::Vector ()
{}
// ------------------------------------------------------------------------------------------------

//! Konstruktor mit Längenangabe
Vector::Vector( unsigned int n )
: vector<double> ( n )
{}
// ------------------------------------------------------------------------------------------------

//! Konstruktor aus Gitter
Vector::Vector( const Grid & g )
: vector<double> ( g.getNumberOfNodes() )
{}
// ------------------------------------------------------------------------------------------------

//! Kopier-Konstruktor
Vector::Vector( const Vector& v )
: vector<double> ( v )
{}
// ------------------------------------------------------------------------------------------------

//! Destruktor
Vector::~Vector ()
{
    // Der Destruktor der Basisklasse (STL-Vektor) wird automatisch aufgerufen
}
// ------------------------------------------------------------------------------------------------

//! Zuweisungsoperator
Vector& Vector::operator=( const Vector& v )
{
    // Verwende den Zuweisungsoperator des STl-Vektors
    (*this).vector<double>::operator=( v );
    return *this;
}
// ------------------------------------------------------------------------------------------------

//! Resize
void Vector::resize( unsigned int n )
{
    // Verwende resize des STl-Vektors
    (*this).vector<double>::resize( n );
}
// ------------------------------------------------------------------------------------------------

//! Setze alle Werte auf \a value
void Vector::setAll( double value )
{
    // size() is eine Funktion des STL-Vektors
    for ( unsigned int i = 0; i < size(); i++ )
    {
        // der Zugriff über den [] Operator stammt ebenfalls von STL-Vektor
        ( *this )[i] = value;
    }
}
// ------------------------------------------------------------------------------------------------

//! Setze alle Werte auf 0
void Vector::setZero ()
{
    setAll(0.);
}
// ------------------------------------------------------------------------------------------------

//! Normiert den Vektor
void Vector::normalize ()
{
    this->operator/=( this->norm_l2() );
}
// ------------------------------------------------------------------------------------------------

//! l2-Norm ausgeben
double Vector::norm_l2 () const
{
    return sqrt ( norm_l2q () );
}
// ------------------------------------------------------------------------------------------------

//! Quadrat der l2-Norm ausgeben
double Vector::norm_l2q () const
{
    double r = 0.0;
    for ( unsigned int i = 0; i < size(); i++ )
    {
        r += ( *this )[i] * ( *this )[i];
    }
    return r;
}
// ------------------------------------------------------------------------------------------------

//! l-unendlich-Norm ausgeben
double Vector::norm_linf () const
{
    double ret = 0.0;
    for ( unsigned i = 0; i < size(); ++i)
    {
        if ( fabs( (*this)[i] ) > ret ) ret = fabs( (*this)[i] );
    }
    return ret;
}
// ------------------------------------------------------------------------------------------------

//! *this = d * v setzen (d reell, v Vektor)
void Vector::setToMultiple( const double d, const Vector& v )
{
    assert( size() == v.size() );

    for ( unsigned int i = 0; i < size(); i++ )
    {
        (*this)[i] = d * v[i];
    }
}
// ------------------------------------------------------------------------------------------------

//! *this += d * v berechnen (d reell, v Vektor)
void Vector::addMultiple( const double d, const Vector& v )
{
    assert( size() == v.size() );

    for ( unsigned int i = 0; i < size(); i++ )
    {
        (*this)[i] += d * v[i];
    }
}
// ------------------------------------------------------------------------------------------------

//! *= Operator: Multipliziert jeden Eintrag mit Faktor r
void Vector::operator*=( const double r )
{
    for ( unsigned int i = 0; i < size(); i++ )
    {
        (*this)[i] *= r;
    }
}
// ------------------------------------------------------------------------------------------------

//! /= Operator: Dividiert jeden Eintrag durch Faktor r
void Vector::operator/=( const double r )
{
    assert ( fabs( r ) > 1e-200 );

    double c = 1.0/r;
    for ( unsigned int i = 0; i < size(); i++ )
    {
        (*this)[i] *= c;
    }
}
// ------------------------------------------------------------------------------------------------

//! += Operator: Addiert einen anderen Vektor
void Vector::operator+=( const Vector& v )
{
    assert( size() == v.size() );

    for ( unsigned int i = 0; i < size(); i++ )
    {
        (*this)[i] += v[i];
    }
}
// ------------------------------------------------------------------------------------------------

//! -= Operator: Subtrahiert einen anderen Vektor
void Vector::operator-=( const Vector& v )
{
    assert( size() == v.size() );

    for ( unsigned int i = 0; i < size(); i++ )
    {
        (*this)[i] -= v[i];
    }
}
// ------------------------------------------------------------------------------------------------

//! * Operator: Berechnet das Skalarprodukt zweier Vektoren
double Vector::operator*( const Vector& v ) const
{
    assert( size() == v.size() );

    double ret = 0.0;
    for ( unsigned int i = 0; i < size(); i++ )
    {
        ret += (*this)[i] * v[i];
    }
    return ret;
}
// ------------------------------------------------------------------------------------------------

//! == Operator: Testet auf Gleichheit zweier Vektoren
bool Vector::operator==( const Vector & alius ) const
{
    double TOL = 1e-15;

    if ( size() != alius.size() ) return false;
    // else
    for ( unsigned int i = 0; i < size(); i++ )
    {
        if ( fabs( (*this)[i] - alius[i] ) > TOL ) return false;
    }
    // else
    return true;
}
// ------------------------------------------------------------------------------------------------

//! != Operator: Testet auf Ungleichheit zweier Vektoren
bool Vector::operator!=( const Vector & alius ) const
{
    return !( *this == alius );
}
// ------------------------------------------------------------------------------------------------

//! Ausgabefunktion mit <<-Operator
ostream& operator<<( ostream& stream, const Vector& vector )
{
    stream.precision(2);
    stream.setf( ios::fixed, ios::floatfield );

    stream << "[";
    if ( vector.size() > 0)  {
        for ( unsigned int i = 0; i < vector.size(); i++ )
        {
            stream << " " << setw(8) << vector[i] << " ";
        }
    }
    stream << "]" << endl;

    stream.precision(6);
    stream.unsetf( ios::fixed );

    return stream;
}
// ------------------------------------------------------------------------------------------------
